import json
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Self

import cattrs
import jax
from jaxtyping import Array, DTypeLike

from lalamo.modules import Decoder, DecoderConfig

__all__ = ["ForeignConfig"]


@dataclass(frozen=True)
class ForeignConfig:
    _converter: ClassVar[cattrs.Converter] = cattrs.Converter()
    _converter.register_structure_hook(int | list[int], lambda v, _: v)

    @property
    @abstractmethod
    def default_precision(self) -> DTypeLike: ...

    @classmethod
    def from_json(cls, json_path: Path | str) -> Self:
        json_path = Path(json_path)
        with open(json_path) as f:
            config = json.load(f)
        return cls._converter.structure(config, cls)

    def to_json(self, json_path: Path | str) -> None:
        json_path = Path(json_path)
        with open(json_path, "w") as f:
            json.dump(self._converter.unstructure(self), f, indent=2)

    def to_decoder_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
    ) -> DecoderConfig:
        raise NotImplementedError

    @classmethod
    def _load_weights(
        cls,
        model: Decoder,
        weights_dict: dict[str, Array],
    ) -> Decoder:
        raise NotImplementedError

    def load_model(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
        weights_dict: dict[str, Array],
    ) -> Decoder:
        config = self.to_decoder_config(context_length, activation_precision, accumulation_precision)
        model = config.random_init(key=jax.random.PRNGKey(0))
        return self._load_weights(model, weights_dict)



from abc import abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from typing import NamedTuple

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float, Int, PRNGKeyArray

from lalamo.modules import Decoder, KVCache

__all__ = [
    "BanTokensPolicy",
    "CompositePolicy",
    "GreedyPolicy",
    "LanguageModel",
    "SamplingPolicy",
    "TemperaturePolicy",
    "TopKPolicy",
    "TopPPolicy",
]


class SamplingPolicy(eqx.Module):
    @abstractmethod
    def process_logits(self, logits: Float[Array, " vocabulary"]) -> Float[Array, " vocabulary"]: ...

    def __call__(self, logits: Float[Array, " vocabulary"], *, key: PRNGKeyArray) -> Int[Array, ""]:
        return jax.random.categorical(key, self.process_logits(logits))


class GreedyPolicy(SamplingPolicy):
    def process_logits(self, logits: Float[Array, " vocabulary"]) -> Float[Array, " vocabulary"]:
        max_logit_value = jnp.max(logits)
        return jnp.where(logits == max_logit_value, 1.0, -jnp.inf)


class TemperaturePolicy(SamplingPolicy):
    temperature: float = eqx.field(static=True)

    def process_logits(self, logits: Float[Array, " vocabulary"]) -> Float[Array, " vocabulary"]:
        return logits / self.temperature


class TopKPolicy(SamplingPolicy):
    k: int = eqx.field(static=True)

    def process_logits(self, logits: Float[Array, " vocabulary"]) -> Float[Array, " vocabulary"]:
        top_k_logits, _ = jax.lax.top_k(logits, self.k)
        min_logit_val = jnp.min(top_k_logits)
        return jnp.where(logits >= min_logit_val, logits, -jnp.inf)


class TopPPolicy(SamplingPolicy):
    p: float = eqx.field(static=True)

    def process_logits(self, logits: Float[Array, " vocabulary"]) -> Float[Array, " vocabulary"]:
        sorted_indices = jnp.argsort(logits, descending=True)
        sorted_logits = logits[sorted_indices]
        cumulative_probs = jnp.cumsum(jax.nn.softmax(sorted_logits))

        to_remove = cumulative_probs > self.p
        to_remove = jnp.roll(to_remove, 1)
        to_remove = to_remove.at[0].set(False)

        return jnp.where(to_remove, -jnp.inf, logits)


class BanTokensPolicy(SamplingPolicy):
    banned_tokens: list[int] = eqx.field(static=True)

    def process_logits(self, logits: Float[Array, " vocabulary"]) -> Float[Array, " vocabulary"]:
        banned_tokens_indices = jnp.asarray(self.banned_tokens, dtype=jnp.int32)
        return logits.at[banned_tokens_indices].set(-jnp.inf)


class CompositePolicy(SamplingPolicy):
    policies: list[SamplingPolicy] = eqx.field(static=True)

    def process_logits(self, logits: Float[Array, " vocabulary"]) -> Float[Array, " vocabulary"]:
        for policy in self.policies:
            logits = policy.process_logits(logits)
        return logits


class PrefillResults(NamedTuple):
    last_token_logits: Float[Array, " vocabulary"]
    last_token_position: Int[Array, ""]
    kv_cache: KVCache


class DecodingState(NamedTuple):
    last_token_logits: Float[Array, " vocabulary"]
    last_token_position: Int[Array, ""]
    kv_cache: KVCache
    stop_flag: Bool[Array, ""]


@dataclass(frozen=True)
class LanguageModel:
    decoder: Decoder

    def _prefill(
        self,
        token_ids: Int[Array, " tokens"],
        length_without_padding: Int[Array, ""] | int | None = None,
        kv_cache_capacity: int | None = None,
    ) -> PrefillResults:
        (num_tokens,) = token_ids.shape
        token_positions = jnp.arange(num_tokens, dtype=jnp.int32)
        if kv_cache_capacity is not None:
            kv_cache = self.decoder.init_static_kv_cache(kv_cache_capacity)
        else:
            kv_cache = None

        decoder_outputs = self.decoder(
            token_ids,
            token_positions,
            kv_cache,
            return_updated_kv_cache=True,
            length_without_padding=length_without_padding,
        )

        if length_without_padding is not None:
            last_logits_index = length_without_padding - 1
        else:
            last_logits_index = num_tokens - 1

        last_token_logits = decoder_outputs.logits[last_logits_index, :]
        last_token_position = jnp.array(last_logits_index, dtype=jnp.int32)

        assert decoder_outputs.updated_kv_cache is not None
        return PrefillResults(
            last_token_logits=last_token_logits,
            last_token_position=last_token_position,
            kv_cache=decoder_outputs.updated_kv_cache,
        )

    def generate(
        self,
        prompt_token_ids: Int[Array, " prompt_tokens"],
        sampling_policy: SamplingPolicy | None = None,
        prompt_length_without_padding: Int[Array, ""] | int | None = None,
        max_output_length: int = 8192,
        eos_token_ids: Int[Array, " eos_tokens"] | None = None,
        *,
        key: PRNGKeyArray | None = None,
    ) -> Int[Array, " response_tokens"]:
        if sampling_policy is None:
            sampling_policy = TemperaturePolicy(temperature=1.0)

        (input_length,) = prompt_token_ids.shape
        prefill_results = self._prefill(
            prompt_token_ids,
            prompt_length_without_padding,
            input_length + max_output_length,
        )

        initial_state = DecodingState(
            prefill_results.last_token_logits,
            prefill_results.last_token_position,
            prefill_results.kv_cache,
            jnp.array(0, dtype=jnp.bool),
        )

        if key is None:
            key = jax.random.PRNGKey(0)
        keys = jax.random.split(key, num=max_output_length)

        def loop_iteration(
            state: DecodingState,
            key: PRNGKeyArray,
        ) -> tuple[DecodingState, Int[Array, ""]]:
            def sample_and_update() -> tuple[DecodingState, Int[Array, ""]]:
                processed_logits = sampling_policy.process_logits(state.last_token_logits)
                next_token_id = jax.random.categorical(key, processed_logits)
                next_token_position = state.last_token_position + 1

                if eos_token_ids is not None:
                    stop_flag = state.stop_flag | jnp.any(next_token_id == eos_token_ids)
                else:
                    stop_flag = state.stop_flag

                decoder_outputs = self.decoder(
                    next_token_id.reshape(1),
                    next_token_position.reshape(1),
                    state.kv_cache,
                    return_updated_kv_cache=True,
                )
                assert decoder_outputs.updated_kv_cache is not None, "updated_kv_cache should not be None"
                new_state = DecodingState(
                    decoder_outputs.logits.squeeze(),
                    next_token_position,
                    decoder_outputs.updated_kv_cache,
                    stop_flag,
                )
                return new_state, next_token_id

            def pad_and_repeat_state() -> tuple[DecodingState, Int[Array, ""]]:
                pad_token = jnp.array(0, dtype=jnp.int32)
                return state, pad_token

            return jax.lax.cond(state.stop_flag, pad_and_repeat_state, sample_and_update)

        _, tokens = jax.lax.scan(loop_iteration, initial_state, keys)

        return tokens

    def stream(
        self,
        prompt_token_ids: Int[Array, " prompt_tokens"],
        sampling_policy: SamplingPolicy | None = None,
        prompt_length_without_padding: Int[Array, ""] | int | None = None,
        max_output_length: int = 8192,
        eos_token_ids: Int[Array, " eos_tokens"] | None = None,
        *,
        key: PRNGKeyArray | None = None,
    ) -> Iterable[Int[Array, ""]]:
        if sampling_policy is None:
            sampling_policy = TemperaturePolicy(temperature=1.0)

        (input_length,) = prompt_token_ids.shape
        prefill_results = self._prefill(
            prompt_token_ids,
            prompt_length_without_padding,
            input_length + max_output_length,
        )

        if key is None:
            key = jax.random.PRNGKey(0)
        keys = jax.random.split(key, num=max_output_length)

        state = DecodingState(
            prefill_results.last_token_logits,
            prefill_results.last_token_position,
            prefill_results.kv_cache,
            jnp.array(0, dtype=jnp.bool),
        )

        for iter_key in keys:
            processed_logits = sampling_policy.process_logits(state.last_token_logits)
            next_token_id = jax.random.categorical(iter_key, processed_logits)

            yield next_token_id

            if eos_token_ids is not None and jnp.any(next_token_id == eos_token_ids):
                return

            next_token_position = state.last_token_position + 1
            decoder_outputs = self.decoder(
                next_token_id.reshape(1),
                next_token_position.reshape(1),
                state.kv_cache,
                return_updated_kv_cache=True,
            )
            assert decoder_outputs.updated_kv_cache is not None, "updated_kv_cache should not be None"
            state = DecodingState(
                decoder_outputs.logits.squeeze(),
                next_token_position,
                decoder_outputs.updated_kv_cache,
                state.stop_flag,
            )



from lalamo.model_import.configs import HFLlamaConfig

from .common import (
    HUGGINFACE_GENERATION_CONFIG_FILE,
    HUGGINGFACE_TOKENIZER_FILES,
    ModelSpec,
    WeightsType,
    huggingface_weight_files,
)

__all__ = ["PLEIAS_MODELS"]

PLEIAS_MODELS = [
    ModelSpec(
        vendor="PleIAs",
        family="Pleias-RAG",
        name="Pleias-RAG-1B",
        size="1B",
        quantization=None,
        repo="PleIAs/Pleias-RAG-1B",
        config_type=HFLlamaConfig,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(1),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
]



from dataclasses import dataclass
from typing import Literal

import jax.numpy as jnp
from jaxtyping import Array, DTypeLike

from lalamo.model_import.configs import ForeignConfig
from lalamo.model_import.loaders import load_huggingface
from lalamo.modules import Decoder

__all__ = [
    "HuggingFaceConfig",
    "AWQQuantizationConfig",
    "GPTQMetaConfig",
    "GPTQQuantizationConfig"
]


@dataclass(frozen=True)
class AWQQuantizationConfig:
    backend: Literal["autoawq"] = "autoawq"
    bits: Literal[4, 8] = 4
    do_fuse: Literal[False] = False
    exllama_config: None = None
    fuse_max_seq_len: None = None
    group_size: int = 128
    modules_to_fuse: None = None
    modules_to_not_convert: None = None
    quant_method: Literal["awq"] = "awq"
    version: Literal["gemm"] = "gemm"
    zero_point: bool = True


@dataclass(frozen=True)
class GPTQMetaConfig:
    damp_auto_increment: float
    damp_percent: float
    mse: float
    quantizer: list[str]
    static_groups: bool
    true_sequential: bool
    uri: str


@dataclass(frozen=True)
class GPTQQuantizationConfig:
    bits: int
    checkpoint_format: str
    desc_act: bool
    group_size: int
    lm_head: bool
    meta: GPTQMetaConfig
    pack_dtype: str
    quant_method: Literal["gptq"]
    sym: bool


@dataclass(frozen=True)
class HuggingFaceConfig(ForeignConfig):
    torch_dtype: Literal["bfloat16", "float16", "float32"]

    @property
    def default_precision(self) -> DTypeLike:
        return jnp.dtype(self.torch_dtype)

    @classmethod
    def _load_weights(
        cls,
        model: Decoder,
        weights_dict: dict[str, Array],
    ) -> Decoder:
        return load_huggingface(model, weights_dict)



from enum import Enum

import jax
import jax.numpy as jnp
from jax import jit
from jaxtyping import Array, Float

__all__ = [
    "Activation",
    "silu",
]


@jit
def silu(x: Float[Array, "*dims"]) -> Float[Array, "*dims"]:
    return x / (1 + jnp.exp(-x))


class Activation(Enum):
    SILU = "silu"
    GELU = "gelu"

    def __call__(self, x: Float[Array, "*dims"]) -> Float[Array, "*dims"]:
        return ACTIVATION_FUNCTIONS[self](x)


ACTIVATION_FUNCTIONS = {
    Activation.SILU: silu,
    Activation.GELU: jax.nn.gelu,
}



from lalamo.model_import.configs import HFLlamaConfig

from .common import (
    HUGGINFACE_GENERATION_CONFIG_FILE,
    HUGGINGFACE_TOKENIZER_FILES,
    ModelSpec,
    WeightsType,
    huggingface_weight_files,
)

__all__ = ["REKA_MODELS"]

REKA_MODELS = [
    ModelSpec(
        vendor="Reka",
        family="Reka-Flash",
        name="Reka-Flash-3.1",
        size="21B",
        quantization=None,
        repo="RekaAI/reka-flash-3.1",
        config_type=HFLlamaConfig,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(9),  # Model has 9 shards
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
] 


from dataclasses import replace

from lalamo.model_import.configs import HFLlamaConfig

from .common import (
    HUGGINFACE_GENERATION_CONFIG_FILE,
    HUGGINGFACE_TOKENIZER_FILES,
    ModelSpec,
    TokenizerFileSpec,
    WeightsType,
    huggingface_weight_files,
)

__all__ = ["LLAMA_MODELS"]

LLAMA31 = [
    ModelSpec(
        vendor="Meta",
        family="Llama-3.1",
        name="Llama-3.1-8B-Instruct",
        size="8B",
        quantization=None,
        repo="meta-llama/Llama-3.1-8B-Instruct",
        config_type=HFLlamaConfig,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(4),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
]


def _tokenizer_files_from_another_repo(repo: str) -> tuple[TokenizerFileSpec, ...]:
    return tuple(
        replace(spec, repo=repo) for spec in (*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE)
    )


LLAMA32 = [
    # LLAMA
    ModelSpec(
        vendor="Meta",
        family="Llama-3.2",
        name="Llama-3.2-1B-Instruct",
        size="1B",
        quantization=None,
        repo="meta-llama/Llama-3.2-1B-Instruct",
        config_type=HFLlamaConfig,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(1),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    # ModelSpec(
    #     vendor="Meta",
    #     family="Llama-3.2",
    #     name="Llama-3.2-1B-Instruct-QLoRA",
    #     size="1B",
    #     quantization=QuantizationMode.UINT4,
    #     repo="meta-llama/Llama-3.2-1B-Instruct-QLORA_INT4_EO8",
    #     config_type=ETLlamaConfig,
    #     config_file_name="params.json",
    #     weights_file_names=("consolidated.00.pth",),
    #     weights_type=WeightsType.TORCH,
    #     tokenizer_files=_tokenizer_files_from_another_repo("meta-llama/Llama-3.2-1B-Instruct"),
    #     use_cases=tuple(),
    # ),
    ModelSpec(
        vendor="Meta",
        family="Llama-3.2",
        name="Llama-3.2-3B-Instruct",
        size="3B",
        quantization=None,
        repo="meta-llama/Llama-3.2-3B-Instruct",
        config_type=HFLlamaConfig,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(2),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    # ModelSpec(
    #     vendor="Meta",
    #     family="Llama-3.2",
    #     name="Llama-3.2-3B-Instruct-QLoRA",
    #     size="3B",
    #     quantization=QuantizationMode.UINT4,
    #     repo="meta-llama/Llama-3.2-3B-Instruct-QLORA_INT4_EO8",
    #     config_type=ETLlamaConfig,
    #     config_file_name="params.json",
    #     weights_file_names=("consolidated.00.pth",),
    #     tokenizer_files=_tokenizer_files_from_another_repo("meta-llama/Llama-3.2-3B-Instruct"),
    #     weights_type=WeightsType.TORCH,
    #     use_cases=tuple(),
    # ),
]

LLAMA_MODELS = LLAMA31 + LLAMA32



from lalamo.model_import.configs import HFQwen2Config

from .common import (
    HUGGINFACE_GENERATION_CONFIG_FILE,
    HUGGINGFACE_TOKENIZER_FILES,
    ModelSpec,
    WeightsType,
    huggingface_weight_files,
)

__all__ = ["DEEPSEEK_MODELS"]

DEEPSEEK_MODELS = [
    ModelSpec(
        vendor="DeepSeek",
        family="R1-Distill-Qwen",
        name="R1-Distill-Qwen-1.5B",
        size="1.5B",
        quantization=None,
        repo="deepseek-ai/DeepSeek-R1-Distill-Qwen-1.5B",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(1),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
]



import einops
import jax.numpy as jnp
import torch.utils.dlpack
from jaxtyping import Array

__all__ = [
    "jax_to_torch",
    "jax_uint4_to_packed_uint8",
    "torch_to_jax",
]


@torch.no_grad()
def _torch_to_jax_bfloat16(tensor: torch.Tensor) -> Array:
    # Credit: https://github.com/jax-ml/ml_dtypes/issues/81#issuecomment-2399636232
    if tensor.dtype != torch.bfloat16:
        raise ValueError("Trying to convert non-bfloat16 tensor to bfloat16")
    intermediate_tensor = tensor.view(torch.uint16)
    return jnp.array(intermediate_tensor).view("bfloat16")


def torch_to_jax(array: torch.Tensor) -> Array:
    array = array.detach().cpu()
    if array.dtype == torch.bfloat16:
        return _torch_to_jax_bfloat16(array)
    return jnp.array(array.numpy())


def jax_to_torch(array: Array) -> torch.Tensor:
    if array.dtype == jnp.bfloat16:
        intermediate_array = array.view(jnp.uint16)
        return torch.utils.dlpack.from_dlpack(intermediate_array).view(torch.bfloat16)
    return torch.utils.dlpack.from_dlpack(array)


def jax_uint4_to_packed_uint8(array: Array) -> Array:
    if array.dtype != jnp.uint4:
        raise ValueError(f"Input array must have dtype jnp.uint4, but got {array.dtype}")

    if not array.shape:
        raise ValueError("Input array cannot be a scalar and must have at least one dimension.")

    *_, last_dim = array.shape
    if last_dim % 2 != 0:
        raise ValueError(f"The last dimension of the input array must be even, but got shape {array.shape}")

    low_nibbles, high_nibbles = einops.rearrange(
        array.astype(jnp.uint8),
        "... (dim_half two) -> two ... dim_half",
        two=2,
    )

    packed = (high_nibbles << 4) | low_nibbles

    return packed.astype(jnp.uint8)



from dataclasses import replace

from lalamo.model_import.configs import HFMistralConfig

from .common import (
    HUGGINFACE_GENERATION_CONFIG_FILE,
    HUGGINGFACE_TOKENIZER_FILES,
    ModelSpec,
    TokenizerFileSpec,
    UseCase,
    WeightsType,
    huggingface_weight_files,
)

__all__ = ["MISTRAL_MODELS"]

CODESTRAL = [
    ModelSpec(
        vendor="Mistral",
        family="Codestral",
        name="Codestral-22B-v0.1",
        size="22B",
        quantization=None,
        repo="mistral-community/Codestral-22B-v0.1",
        config_type=HFMistralConfig,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(9),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=(UseCase.CODE,),
    ),
]


def _tokenizer_files_from_another_repo(repo: str) -> tuple[TokenizerFileSpec, ...]:
    return tuple(
        replace(spec, repo=repo) for spec in (*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE)
    )


DEVSTRAL = [
    ModelSpec(
        vendor="Mistral",
        family="Devstral",
        name="Devstral-Small-2505",
        size="24B",
        quantization=None,
        repo="mistralai/Devstral-Small-2505",
        config_type=HFMistralConfig,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(10),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=_tokenizer_files_from_another_repo("mistralai/Mistral-Small-3.1-24B-Base-2503"),
        use_cases=(UseCase.CODE,),
    ),
]


MISTRAL_MODELS = CODESTRAL + DEVSTRAL



import jax.numpy as jnp
from einops import rearrange
from jaxtyping import Array

from lalamo.common import ParameterPath
from lalamo.modules import (
    MLP,
    Attention,
    Decoder,
    DecoderLayer,
    FullPrecisionLinear,
    GroupQuantizedLinear,
    LinearBase,
    RMSNorm,
    TiedEmbedding,
    UntiedEmbedding,
)
from lalamo.quantization import QuantizationMode

from .common import load_parameters

__all__ = ["load_huggingface"]


AWQ_REVERSE_ORDER = jnp.array([0, 4, 1, 5, 2, 6, 3, 7], dtype=jnp.int32)


def _reverse_uint4_awq_order(array: Array) -> Array:
    """Reverses the AWQ packing order to get the logical order of channels for INT4."""
    pack_factor = 32 // 4
    *_, last_dim = array.shape
    if last_dim % pack_factor != 0:
        return array

    array_reshaped = rearrange(array, "... (group pack_factor) -> ... group pack_factor", pack_factor=pack_factor)
    array_reordered = array_reshaped[..., AWQ_REVERSE_ORDER]
    return rearrange(array_reordered, "... group pack_factor -> ... (group pack_factor)")


def unpack_int32(packed_weights: Array, mode: QuantizationMode) -> Array:
    assert packed_weights.dtype == jnp.int32, (
        f"Expected packed_weights to be of dtype jnp.int32, got {packed_weights.dtype}"
    )
    assert 32 % mode.bits == 0

    shifts = jnp.arange(0, 32, mode.bits)
    mask = (2**mode.bits) - 1
    unpacked = jnp.bitwise_and(jnp.right_shift(packed_weights[:, :, None], shifts[None, None, :]), mask)
    unpacked = rearrange(
        unpacked,
        "out_channels packed_groups packed_values -> out_channels (packed_groups packed_values)",
    )

    return unpacked


def _process_quantized_tensors(
    qweights: Array,
    qzeros: Array,
    scales: Array,
    module: GroupQuantizedLinear,
) -> tuple[Array, Array, Array]:
    """Unpacks, recenters, transposes, and casts quantized tensors to the correct dtype."""
    mode = module.config.weight_quantization_mode
    assert qweights.dtype == jnp.int32
    unpacked_weights = unpack_int32(qweights, mode)
    if mode == QuantizationMode.UINT4:
        unpacked_weights = _reverse_uint4_awq_order(unpacked_weights)

    assert qzeros.dtype == jnp.int32
    unpacked_zero_points = unpack_int32(qzeros, mode)
    if mode == QuantizationMode.UINT4:
        unpacked_zero_points = _reverse_uint4_awq_order(unpacked_zero_points)

    weights = unpacked_weights.astype(module.config.activation_precision)
    zero_points = unpacked_zero_points.astype(module.config.activation_precision)
    processed_scales = scales.astype(module.config.activation_precision)

    return weights.transpose(), zero_points.transpose(), processed_scales.transpose()


def _fuse_full_precision_weights(
    weights_dict: dict[str, Array],
    path: ParameterPath,
    sublayers_to_fuse: list[str] | None,
) -> Array:
    if sublayers_to_fuse is None:
        return weights_dict[path / "weight"]

    weights = [weights_dict[path / layer_name / "weight"] for layer_name in sublayers_to_fuse]
    return jnp.concatenate(weights, axis=0)


def _fuse_quantized_weights(
    weights_dict: dict[str, Array],
    path: ParameterPath,
    sublayers_to_fuse: list[str] | None,
) -> tuple[Array, Array, Array]:
    # Note that AWQ quantized weights are stored transposed relative to full-precision weights

    if sublayers_to_fuse is None:
        qweights = weights_dict[path / "qweight"]
        qzeros = weights_dict[path / "qzeros"]
        scales = weights_dict[path / "scales"]
        return qweights, qzeros, scales

    qweights = [weights_dict[path / layer_name / "qweight"] for layer_name in sublayers_to_fuse]
    qzeros = [weights_dict[path / layer_name / "qzeros"] for layer_name in sublayers_to_fuse]
    scales = [weights_dict[path / layer_name / "scales"] for layer_name in sublayers_to_fuse]

    fused_qweights = jnp.concatenate(qweights, axis=1)
    fused_qzeros = jnp.concatenate(qzeros, axis=1)
    fused_scales = jnp.concatenate(scales, axis=1)

    return fused_qweights, fused_qzeros, fused_scales


def load_linear(
    module: LinearBase,
    weights_dict: dict[str, Array],
    path: ParameterPath,
    sublayers_to_fuse: list[str] | None = None,
) -> LinearBase:
    """Loads a linear layer, optionally fusing weights from sublayers."""
    if not module.has_biases:
        if sublayers_to_fuse:
            paths_to_check = [path / proj / "bias" for proj in sublayers_to_fuse]
        else:
            paths_to_check = path / "bias"
        for p in paths_to_check:
            if p in weights_dict:
                raise ValueError(f"Bias tensor found at {p} but module does not support it.")
        bias = None
    elif sublayers_to_fuse is None:
        bias = weights_dict[path / "bias"]
    else:
        bias = jnp.concatenate(
            [weights_dict[path / proj_name / "bias"] for proj_name in sublayers_to_fuse],
            axis=0,
        )

    if isinstance(module, FullPrecisionLinear):
        weights = _fuse_full_precision_weights(weights_dict, path, sublayers_to_fuse)
        return load_parameters(lambda m: (m.weights, m.biases), module, (weights, bias))

    if isinstance(module, GroupQuantizedLinear):
        qweights, qzeros, scales = _fuse_quantized_weights(weights_dict, path, sublayers_to_fuse)

        weights, zero_points, scales = _process_quantized_tensors(
            qweights,
            qzeros,
            scales,
            module,
        )

        return load_parameters(
            lambda m: (m.weights, m.scales, m.zero_points, m.biases),
            module,
            (weights, scales, zero_points, bias),
        )

    raise TypeError(f"Unsupported module type for loading: {type(module)}")


def load_mlp(module: MLP, weights_dict: dict[str, Array], path: ParameterPath) -> MLP:
    up_projection = load_linear(module.up_projection, weights_dict, path, sublayers_to_fuse=["up_proj", "gate_proj"])
    down_projection = load_linear(module.down_projection, weights_dict, path / "down_proj")
    return load_parameters(lambda m: (m.up_projection, m.down_projection), module, (up_projection, down_projection))


def load_rmsnorm(
    module: RMSNorm,
    weights_dict: dict[str, Array],
    path: ParameterPath,
) -> RMSNorm:
    scales = weights_dict[path / "weight"]
    return load_parameters(lambda m: (m.scales,), module, (scales,))


def load_attention(
    module: Attention,
    weights_dict: dict[str, Array],
    path: ParameterPath,
) -> Attention:
    qkv_projection = load_linear(
        module.qkv_projection,
        weights_dict,
        path,
        sublayers_to_fuse=["q_proj", "k_proj", "v_proj"],
    )
    out_projection = load_linear(module.out_projection, weights_dict, path / "o_proj")

    if module.query_norm is not None:
        query_norm = load_rmsnorm(module.query_norm, weights_dict, path / "q_norm")
    else:
        query_norm = None

    if module.key_norm is not None:
        key_norm = load_rmsnorm(module.key_norm, weights_dict, path / "k_norm")
    else:
        key_norm = None

    return load_parameters(
        lambda m: (m.qkv_projection, m.out_projection, m.query_norm, m.key_norm),
        module,
        (qkv_projection, out_projection, query_norm, key_norm),
    )


def load_decoder_layer(
    module: DecoderLayer,
    weights_dict: dict[str, Array],
    path: ParameterPath,
) -> DecoderLayer:
    pre_attention_norm = load_rmsnorm(
        module.pre_attention_norm,
        weights_dict,
        path / "input_layernorm",
    )
    attention = load_attention(module.attention, weights_dict, path / "self_attn")
    if module.post_attention_norm is not None:
        post_attention_norm = load_rmsnorm(
            module.post_attention_norm,
            weights_dict,
            path / "post_attention_layernorm",
        )

        pre_mlp_norm = load_rmsnorm(
            module.pre_mlp_norm,
            weights_dict,
            path / "pre_feedforward_layernorm",
        )
    else:
        post_attention_norm = None

        pre_mlp_norm = load_rmsnorm(
            module.pre_mlp_norm,
            weights_dict,
            path / "post_attention_layernorm",
        )

    mlp = load_mlp(module.mlp, weights_dict, path / "mlp")
    if module.post_mlp_norm is not None:
        post_mlp_norm = load_rmsnorm(
            module.post_mlp_norm,
            weights_dict,
            path / "post_feedforward_layernorm",
        )
    else:
        post_mlp_norm = None
    return load_parameters(
        lambda m: (m.pre_attention_norm, m.attention, m.post_attention_norm, m.pre_mlp_norm, m.mlp, m.post_mlp_norm),
        module,
        (pre_attention_norm, attention, post_attention_norm, pre_mlp_norm, mlp, post_mlp_norm),
    )


def load_tied_embedding(
    module: TiedEmbedding,
    weights_dict: dict[str, Array],
    decoder_path: ParameterPath,
) -> TiedEmbedding:
    weights = weights_dict[decoder_path / "embed_tokens" / "weight"]
    return load_parameters(lambda m: (m.weights,), module, (weights,))


def load_untied_embedding(
    module: UntiedEmbedding,
    weights_dict: dict[str, Array],
    decoder_path: ParameterPath,
    lm_head_path: ParameterPath,
) -> UntiedEmbedding:
    input_weights = weights_dict[decoder_path / "embed_tokens" / "weight"]
    output_weights = weights_dict[lm_head_path / "weight"]
    return load_parameters(lambda m: (m.input_weights, m.output_weights), module, (input_weights, output_weights))


def load_huggingface(
    module: Decoder,
    weights_dict: dict[str, Array],
) -> Decoder:
    if any(key.startswith("language_model.") for key in weights_dict):
        base_path = ParameterPath("language_model")
    else:
        base_path = ParameterPath()

    decoder_path = base_path / "model"
    lm_head_path = base_path / "lm_head"

    if isinstance(module.embedding, TiedEmbedding):
        embedding = load_tied_embedding(module.embedding, weights_dict, decoder_path)
    elif isinstance(module.embedding, UntiedEmbedding):
        embedding = load_untied_embedding(module.embedding, weights_dict, decoder_path, lm_head_path)
    else:
        raise TypeError(f"Unsupported embedding type: {type(module.embedding)}")
    decoder_layers = tuple(
        load_decoder_layer(layer, weights_dict, decoder_path / "layers" / i) for i, layer in enumerate(module.layers)
    )
    output_norm = load_rmsnorm(module.output_norm, weights_dict, decoder_path / "norm")
    return load_parameters(
        lambda m: (m.embedding, m.layers, m.output_norm),
        module,
        (embedding, decoder_layers, output_norm),
    )



from .activations import Activation
from .attention import Attention, AttentionConfig
from .common import WeightLayout, config_converter
from .decoder import Decoder, DecoderActivationTrace, DecoderConfig, DecoderResult
from .decoder_layer import DecoderLayer, DecoderLayerActivationTrace, DecoderLayerConfig, DecoderLayerResult
from .embedding import (
    EmbeddingBase,
    EmbeddingConfig,
    QuantizedTiedEmbedding,
    QuantizedTiedEmbeddingConfig,
    TiedEmbedding,
    TiedEmbeddingConfig,
    UntiedEmbedding,
    UntiedEmbeddingConfig,
)
from .kv_cache import DynamicKVCacheLayer, KVCache, KVCacheLayer, StaticKVCacheLayer
from .linear import (
    FullPrecisionLinear,
    FullPrecisionLinearConfig,
    GroupQuantizedLinear,
    GroupQuantizedLinearConfig,
    LinearBase,
    LinearConfig,
    QLoRALinear,
    QLoRALinearConfig,
)
from .mlp import MLP, MLPConfig
from .normalization import RMSNorm, RMSNormConfig, UpcastMode
from .rope import (
    LinearScalingRoPEConfig,
    LlamaRoPEConfig,
    PositionalEmbeddings,
    RoPE,
    RoPEConfig,
    UnscaledRoPEConfig,
    YARNRoPEConfig,
)

__all__ = [
    "MLP",
    "Activation",
    "Attention",
    "AttentionConfig",
    "Decoder",
    "DecoderActivationTrace",
    "DecoderConfig",
    "DecoderLayer",
    "DecoderLayerActivationTrace",
    "DecoderLayerConfig",
    "DecoderLayerResult",
    "DecoderResult",
    "DynamicKVCacheLayer",
    "EmbeddingBase",
    "EmbeddingConfig",
    "FullPrecisionLinear",
    "FullPrecisionLinearConfig",
    "GroupQuantizedLinear",
    "GroupQuantizedLinearConfig",
    "KVCache",
    "KVCacheLayer",
    "LinearBase",
    "LinearConfig",
    "LinearScalingRoPEConfig",
    "LlamaRoPEConfig",
    "MLPConfig",
    "PositionalEmbeddings",
    "QLoRALinear",
    "QLoRALinearConfig",
    "QuantizedTiedEmbedding",
    "QuantizedTiedEmbeddingConfig",
    "RMSNorm",
    "RMSNormConfig",
    "RoPE",
    "RoPEConfig",
    "StaticKVCacheLayer",
    "TiedEmbedding",
    "TiedEmbeddingConfig",
    "UnscaledRoPEConfig",
    "UntiedEmbedding",
    "UntiedEmbeddingConfig",
    "UpcastMode",
    "WeightLayout",
    "YARNRoPEConfig",
    "config_converter",
]



from lalamo.model_import.configs import HFQwen2Config, HFQwen3Config
from lalamo.quantization import QuantizationMode

from .common import (
    HUGGINFACE_GENERATION_CONFIG_FILE,
    HUGGINGFACE_TOKENIZER_FILES,
    ModelSpec,
    UseCase,
    WeightsType,
    huggingface_weight_files,
)

__all__ = ["QWEN_MODELS"]


QWEN25 = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-0.5B-Instruct",
        size="0.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-0.5B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(1),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-1.5B-Instruct",
        size="1.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-1.5B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(1),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-3B-Instruct",
        size="3B",
        quantization=None,
        repo="Qwen/Qwen2.5-3B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(2),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-7B-Instruct",
        size="7B",
        quantization=None,
        repo="Qwen/Qwen2.5-7B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(4),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-14B-Instruct",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen2.5-14B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(8),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-32B-Instruct",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen2.5-32B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(17),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
]


QWEN25_CODER = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-0.5B-Instruct",
        size="0.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-0.5B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(1),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-1.5B-Instruct",
        size="1.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-1.5B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(1),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-3B-Instruct",
        size="3B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-3B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(2),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-7B-Instruct",
        size="7B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-7B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(4),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-14B-Instruct",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-14B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(6),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-32B-Instruct",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-32B-Instruct",
        config_type=HFQwen2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(14),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=(UseCase.CODE,),
    ),
]


QWEN3 = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-0.6B",
        size="0.6B",
        quantization=None,
        repo="Qwen/Qwen3-0.6B",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(1),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-1.7B",
        size="1.7B",
        quantization=None,
        repo="Qwen/Qwen3-1.7B",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(2),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-4B",
        size="4B",
        quantization=None,
        repo="Qwen/Qwen3-4B",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(3),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-4B-AWQ",
        size="4B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-4B-AWQ",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(1),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-8B",
        size="8B",
        quantization=None,
        repo="Qwen/Qwen3-8B",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(5),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-8B-AWQ",
        size="8B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-8B-AWQ",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(2),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-14B",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen3-14B",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(8),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-14B-AWQ",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen3-14B-AWQ",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(2),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-32B",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen3-32B",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(17),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-32B-AWQ",
        size="32B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-32B-AWQ",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(4),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
]


QWEN_MODELS = QWEN25 + QWEN25_CODER + QWEN3



from dataclasses import dataclass

import jax
from jaxtyping import Array, DTypeLike, Float, PRNGKeyArray

from lalamo.common import ParameterDict

from .activations import Activation
from .common import LalamoModule, WeightLayout
from .linear import LinearBase, LinearConfig

__all__ = ["MLP", "MLPConfig"]


@dataclass(frozen=True)
class MLPConfig:
    linear_config: LinearConfig
    activation: Activation

    def random_init(self, model_dim: int, hidden_dim: int, *, key: PRNGKeyArray) -> "MLP":
        up_key, down_key = jax.random.split(key)
        return MLP(
            self,
            up_projection=self.linear_config.random_init(
                model_dim,
                (hidden_dim, hidden_dim),
                has_biases=False,
                key=up_key,
            ),
            down_projection=self.linear_config.random_init(
                hidden_dim,
                (model_dim,),
                has_biases=False,
                key=down_key,
            ),
        )


class MLP(LalamoModule):
    up_projection: LinearBase
    down_projection: LinearBase

    @property
    def activation_precision(self) -> DTypeLike:
        return self.up_projection.activation_precision

    @property
    def model_dim(self) -> int:
        return self.up_projection.input_dim

    @property
    def hidden_dim(self) -> int:
        return self.down_projection.input_dim

    def __post_init__(self) -> None:
        up_output_dim, gate_output_dim = self.up_projection.output_dims
        if up_output_dim != gate_output_dim:
            raise ValueError(
                f"Up projection output dimension {up_output_dim} does not match"
                f" the gate output dimension {gate_output_dim}",
            )
        (down_output_dim,) = self.down_projection.output_dims
        if self.up_projection.input_dim != down_output_dim:
            raise ValueError(
                f"Down projection input dimension {down_output_dim} does not match"
                f" the up projection output dimension {self.up_projection.input_dim}",
            )

    def __call__(self, inputs: Float[Array, " channels"]) -> Float[Array, " channels"]:
        up_proj, gate = self.up_projection(inputs)
        gate = self.config.activation(gate)
        (result,) = self.down_projection(up_proj * gate)
        return result

    def export_weights(self, weight_layout: WeightLayout = WeightLayout.AUTO) -> ParameterDict:
        return ParameterDict(
            up_projection=self.up_projection.export_weights(weight_layout),
            down_projection=self.down_projection.export_weights(weight_layout),
        )



from lalamo.model_import.configs import HFGemma2Config, HFGemma3Config, HFGemma3TextConfig

from .common import (
    HUGGINFACE_GENERATION_CONFIG_FILE,
    HUGGINGFACE_TOKENIZER_FILES,
    ModelSpec,
    WeightsType,
    huggingface_weight_files,
)

__all__ = ["GEMMA_MODELS"]

GEMMA2 = [
    ModelSpec(
        vendor="Google",
        family="Gemma-2",
        name="Gemma-2-2B-Instruct",
        size="2B",
        quantization=None,
        repo="google/gemma-2-2b-it",
        config_type=HFGemma2Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(2),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
]

GEMMA3 = [
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-1B-Instruct",
        size="1B",
        quantization=None,
        repo="google/gemma-3-1b-it",
        config_type=HFGemma3TextConfig,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(1),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-4B-Instruct",
        size="4B",
        quantization=None,
        repo="google/gemma-3-4b-it",
        config_type=HFGemma3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(2),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Google",
        family="Gemma-3",
        name="Gemma-3-27B-Instruct",
        size="27B",
        quantization=None,
        repo="google/gemma-3-27b-it",
        config_type=HFGemma3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(12),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, HUGGINFACE_GENERATION_CONFIG_FILE),
        use_cases=tuple(),
    ),
]


GEMMA_MODELS = GEMMA2 + GEMMA3



from textwrap import dedent

import jax.numpy as jnp
import pytest
from attr import dataclass
from jax import jit, vmap
from jaxtyping import Array, Int
from transformers import AutoTokenizer
from transformers.tokenization_utils import PreTrainedTokenizer

from lalamo.language_model import LanguageModel
from lalamo.model_import import REPO_TO_MODEL, import_model


@dataclass
class GenerationInput:
    prompt: str
    token_ids: Int[Array, " tokens"]


@pytest.fixture
def generation_input(tokenizer: PreTrainedTokenizer) -> GenerationInput:
    prompt = dedent(
        """
        <|im_start|>user
        hello<|im_end|>
        <|im_start|>assistant
        <think>\n\n</think>\n\n
    """.lstrip(),
    )
    tokens = jnp.array(tokenizer.encode(prompt))
    return GenerationInput(prompt, tokens)


@pytest.fixture
def another_generation_input(tokenizer: PreTrainedTokenizer) -> GenerationInput:
    prompt = dedent(
        """
        <|im_start|>user
        How are you?<|im_end|>
        <|im_start|>assistant
        <think>\n\n</think>\n\n
    """.lstrip(),
    )
    tokens = jnp.array(tokenizer.encode(prompt))
    return GenerationInput(prompt, tokens)


@pytest.fixture
def language_model() -> LanguageModel:
    decoder = import_model(REPO_TO_MODEL["Qwen/Qwen3-0.6B"]).model
    return LanguageModel(decoder)


@pytest.fixture
def tokenizer() -> PreTrainedTokenizer:
    return AutoTokenizer.from_pretrained("Qwen/Qwen3-0.6B")


def test_eager_generation(
    language_model: LanguageModel,
    tokenizer: PreTrainedTokenizer,
    generation_input: GenerationInput,
) -> None:
    response_token_ids = language_model.generate(generation_input.token_ids, max_output_length=32)
    response_text = tokenizer.decode(response_token_ids)
    assert "<|im_end|>" in response_text


def test_padding(language_model: LanguageModel, tokenizer: PreTrainedTokenizer) -> None:
    prompt = dedent(
        """
        <|im_start|>user
        Talk about elephants<|im_end|>
        <|im_start|>assistant
        <think>\n\n</think>\n\n
    """.lstrip(),
    )
    token_ids = jnp.array(tokenizer.encode(prompt))

    length = jnp.array(0, dtype=jnp.int32)
    response_token_ids = language_model.generate(
        token_ids,
        prompt_length_without_padding=length,
        max_output_length=32,
    )
    response_text = tokenizer.decode(response_token_ids)
    assert "elephants" not in response_text.lower()

    response_token_ids = language_model.generate(
        token_ids,
        prompt_length_without_padding=token_ids.size,
        max_output_length=32,
    )
    response_text = tokenizer.decode(response_token_ids)
    assert "elephants" in response_text.lower()


def test_jit_generation(
    language_model: LanguageModel,
    tokenizer: PreTrainedTokenizer,
    generation_input: GenerationInput,
) -> None:
    response_token_ids = jit(language_model.generate, static_argnames=["max_output_length"])(
        generation_input.token_ids,
        max_output_length=32,
    )
    response_text = tokenizer.decode(response_token_ids)
    assert "<|im_end|>" in response_text


def test_batch_generation(
    language_model: LanguageModel,
    tokenizer: PreTrainedTokenizer,
    generation_input: GenerationInput,
    another_generation_input: GenerationInput,
) -> None:
    def generate_fn(token_ids: Int[Array, " tokens"], sequence_length: Int[Array, ""]) -> Int[Array, " tokens"]:
        return language_model.generate(
            token_ids,
            prompt_length_without_padding=sequence_length,
            max_output_length=32,
        )

    batched_generate_fn = jit(vmap(generate_fn))

    inputs = [generation_input, another_generation_input]
    pad_token_id = 0

    max_len = max(inp.token_ids.size for inp in inputs)

    batched_prompt_lengths = jnp.array([inp.token_ids.size for inp in inputs])
    padded_token_ids = jnp.array(
        [
            jnp.pad(
                inp.token_ids,
                (0, max_len - inp.token_ids.size),
                constant_values=pad_token_id,
            )
            for inp in inputs
        ],
    )

    response_token_ids = batched_generate_fn(padded_token_ids, batched_prompt_lengths)
    for ids in response_token_ids:
        response_text = tokenizer.decode(ids)
        assert "<|im_end|>" in response_text


def test_streaming_generation(
    language_model: LanguageModel,
    tokenizer: PreTrainedTokenizer,
    generation_input: GenerationInput,
) -> None:
    token_stream = language_model.stream(generation_input.token_ids, max_output_length=32)
    response_token_ids = jnp.array(list(token_stream))
    response_text = tokenizer.decode(response_token_ids)
    assert "<|im_end|>" in response_text


def test_streaming_vs_eager_consistency(
    language_model: LanguageModel,
    generation_input: GenerationInput,
) -> None:
    eager_token_ids = language_model.generate(
        generation_input.token_ids,
        max_output_length=32,
        eos_token_ids=jnp.array([-1]),  # Never stop.
    )

    streaming_token_generator = language_model.stream(
        generation_input.token_ids,
        max_output_length=32,
        eos_token_ids=jnp.array([-1]),  # Never stop.
    )
    streaming_token_ids = jnp.array(list(streaming_token_generator))

    assert jnp.array_equal(eager_token_ids, streaming_token_ids)



from lalamo.model_import import REPO_TO_MODEL, ModelSpec, import_model
from lalamo.modules import Decoder

__version__ = "0.2.5"

__all__ = [
    "REPO_TO_MODEL",
    "Decoder",
    "ModelSpec",
    "import_model",
]



from lalamo.model_import.configs import HFQwen3Config

from .common import HUGGINGFACE_TOKENIZER_FILES, ModelSpec, TokenizerFileSpec, WeightsType, huggingface_weight_files

__all__ = ["POLARIS_MODELS"]

POLARIS_MODELS = [
    ModelSpec(
        vendor="POLARIS-Project",
        family="Polaris-Preview",
        name="Polaris-4B-Preview",
        size="4B",
        quantization=None,
        repo="POLARIS-Project/Polaris-4B-Preview",
        config_type=HFQwen3Config,
        config_file_name="config.json",
        weights_file_names=huggingface_weight_files(2),
        weights_type=WeightsType.SAFETENSORS,
        tokenizer_files=(*HUGGINGFACE_TOKENIZER_FILES, TokenizerFileSpec(repo=None, filename="chat_template.jinja")),
        use_cases=tuple(),
    ),
]



from dataclasses import dataclass
from typing import Literal

from jaxtyping import DTypeLike

from lalamo.modules import (
    Activation,
    AttentionConfig,
    DecoderConfig,
    DecoderLayerConfig,
    FullPrecisionLinearConfig,
    GroupQuantizedLinearConfig,
    MLPConfig,
    RMSNormConfig,
    TiedEmbeddingConfig,
    UnscaledRoPEConfig,
    UntiedEmbeddingConfig,
    UpcastMode,
)
from lalamo.quantization import QuantizationMode

from .common import AWQQuantizationConfig, GPTQQuantizationConfig, HuggingFaceConfig

__all__ = ["HFQwen2Config"]


@dataclass(frozen=True)
class HFQwen2Config(HuggingFaceConfig):
    architectures: list[Literal["Qwen2ForCausalLM"]]
    attention_dropout: float
    bos_token_id: int | list[int]
    eos_token_id: int | list[int]
    hidden_act: Literal["silu"]
    hidden_size: int
    initializer_range: float
    intermediate_size: int
    max_position_embeddings: int
    max_window_layers: int
    model_type: Literal["qwen2"]
    num_attention_heads: int
    num_hidden_layers: int
    num_key_value_heads: int
    rms_norm_eps: float
    rope_theta: float
    sliding_window: int
    tie_word_embeddings: bool
    transformers_version: str
    use_cache: bool
    use_sliding_window: bool
    vocab_size: int

    quantization_config: AWQQuantizationConfig | GPTQQuantizationConfig | None = None

    def _get_sliding_window_sizes(self) -> list[int | None]:
        if not self.use_sliding_window:
            return [None] * self.num_hidden_layers

        sliding_window_sizes = []
        for i in range(self.num_hidden_layers):
            if i < self.max_window_layers:
                sliding_window_sizes.append(self.sliding_window)
            else:
                sliding_window_sizes.append(None)
        return sliding_window_sizes

    def to_decoder_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
    ) -> DecoderConfig:
        if self.tie_word_embeddings:
            embedding_config = TiedEmbeddingConfig(
                input_scale=None,
                logits_soft_cap=None,
                precision=activation_precision,
            )
        else:
            embedding_config = UntiedEmbeddingConfig(
                input_scale=None,
                logits_soft_cap=None,
                precision=activation_precision,
            )
        rope_config = UnscaledRoPEConfig(
            precision=activation_precision,
            base=self.rope_theta,
            max_sequence_length=self.max_position_embeddings,
        )
        rmsnorm_config = RMSNormConfig(
            scale_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.rms_norm_eps,
            scale_offset=None,
            upcast_mode=UpcastMode.ONLY_NORMALIZATION,
        )
        if self.quantization_config is None:
            linear_config = FullPrecisionLinearConfig(
                precision=activation_precision,
            )
        else:
            linear_config = GroupQuantizedLinearConfig(
                group_size=self.quantization_config.group_size,
                weight_quantization_mode=QuantizationMode.from_num_bits(self.quantization_config.bits),
                activation_quantization_mode=None,
                activation_precision=activation_precision,
            )
        attention_config = AttentionConfig(
            qkv_projection_config=linear_config,
            out_projection_config=linear_config,
            query_norm_config=None,
            key_norm_config=None,
            logit_soft_cap=None,
            has_qkv_biases=True,
            has_out_biases=False,
        )
        mlp_config = MLPConfig(
            linear_config=linear_config,
            activation=Activation.SILU,
        )
        decoder_layer_config = DecoderLayerConfig(
            pre_attention_norm_config=rmsnorm_config,
            attention_config=attention_config,
            post_attention_norm_config=None,
            pre_mlp_norm_config=rmsnorm_config,
            mlp_config=mlp_config,
            post_mlp_norm_config=None,
        )
        return DecoderConfig(
            embedding_config=embedding_config,
            global_rope_config=rope_config,
            local_rope_config=None,
            layer_config=decoder_layer_config,
            output_norm_config=rmsnorm_config,
            vocab_size=self.vocab_size,
            model_dim=self.hidden_size,
            hidden_dim=self.intermediate_size,
            num_heads=self.num_attention_heads,
            num_groups=self.num_key_value_heads,
            head_dim=self.hidden_size // self.num_attention_heads,
            attention_scale=None,
            num_layers=self.num_hidden_layers,
            sliding_window_sizes=tuple(self._get_sliding_window_sizes()),
            context_length=context_length or self.max_position_embeddings,
        )



from dataclasses import dataclass

import equinox as eqx
import jax
from jax import vmap
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterDict

from .common import AttentionType, LalamoModule, WeightLayout
from .decoder_layer import DecoderLayer, DecoderLayerConfig, DecoderLayerResult
from .embedding import EmbeddingBase, EmbeddingConfig
from .kv_cache import KVCache
from .normalization import RMSNorm, RMSNormConfig
from .rope import PositionalEmbeddings, RoPE, RoPEConfig

__all__ = [
    "Decoder",
    "DecoderActivationTrace",
    "DecoderConfig",
    "DecoderResult",
]


class DecoderActivationTrace(eqx.Module):
    token_ids: Int[Array, " suffix_tokens"]
    token_positions: Int[Array, " suffix_tokens"]
    kv_cache: KVCache | None

    local_positional_embeddings: PositionalEmbeddings
    global_positional_embeddings: PositionalEmbeddings

    layer_results: tuple[DecoderLayerResult, ...]

    output_norm: Float[Array, "suffix_tokens channels"]

    def export(self) -> ParameterDict:
        result = ParameterDict(
            token_ids=self.token_ids,
            token_positions=self.token_positions,
            local_positional_embeddings=self.local_positional_embeddings.export(),
            global_positional_embeddings=self.global_positional_embeddings.export(),
            layer_results=[layer_result.export() for layer_result in self.layer_results],
            output_norm=self.output_norm,
        )
        if self.kv_cache is not None:
            result["kv_cache"] = [kv_cache_layer_slice.export() for kv_cache_layer_slice in self.kv_cache]
        return result


class DecoderResult(eqx.Module):
    logits: Float[Array, "suffix_tokens channels"]
    updated_kv_cache: KVCache | None = None
    activation_trace: DecoderActivationTrace | None = None

    def export(self) -> ParameterDict:
        result = ParameterDict(
            logits=self.logits,
        )
        if self.updated_kv_cache is not None:
            result["updated_kv_cache"] = [
                kv_cache_layer_slice.export() for kv_cache_layer_slice in self.updated_kv_cache
            ]
        if self.activation_trace is not None:
            result["activation_trace"] = self.activation_trace.export()
        return result


@dataclass(frozen=True)
class DecoderConfig:
    embedding_config: EmbeddingConfig
    global_rope_config: RoPEConfig
    local_rope_config: RoPEConfig | None
    layer_config: DecoderLayerConfig
    output_norm_config: RMSNormConfig

    vocab_size: int
    model_dim: int
    hidden_dim: int
    num_heads: int
    num_groups: int
    head_dim: int
    attention_scale: float | None
    num_layers: int
    sliding_window_sizes: tuple[int | None, ...] | None
    context_length: int

    def __post_init__(self) -> None:
        if self.local_rope_config is not None and self.sliding_window_sizes is None:
            raise ValueError("Sliding window sizes must be provided when using local RoPE")
        if self.sliding_window_sizes is None:
            return
        if len(self.sliding_window_sizes) != self.num_layers:
            raise ValueError(
                f"Number of sliding window sizes {len(self.sliding_window_sizes)} does not match"
                f" the number of layers {self.num_layers}",
            )

    def random_init(
        self,
        *,
        key: PRNGKeyArray,
    ) -> "Decoder":
        embedding_key, layers_key = jax.random.split(key)
        embedding = self.embedding_config.random_init(
            vocab_size=self.vocab_size,
            model_dim=self.model_dim,
            key=embedding_key,
        )
        global_rope = self.global_rope_config.init(
            head_dim=self.head_dim,
            num_timesteps=self.context_length,
        )

        if self.local_rope_config:
            assert self.sliding_window_sizes is not None
            max_sliding_window_size = max(
                window_size for window_size in self.sliding_window_sizes if window_size is not None
            )
            local_rope = self.local_rope_config.init(
                head_dim=self.head_dim,
                num_timesteps=max(max_sliding_window_size, self.context_length),
            )
        else:
            local_rope = None

        if self.sliding_window_sizes is None:
            sliding_window_sizes = [None] * self.num_layers
        else:
            sliding_window_sizes = self.sliding_window_sizes
        layers_keys = jax.random.split(layers_key, self.num_layers)
        layers = tuple(
            self.layer_config.random_init(
                model_dim=self.model_dim,
                hidden_dim=self.hidden_dim,
                num_heads=self.num_heads,
                num_groups=self.num_groups,
                head_dim=self.head_dim,
                attention_scale=self.attention_scale,
                sliding_window_size=sliding_window_size,
                key=key,
            )
            for sliding_window_size, key in zip(sliding_window_sizes, layers_keys, strict=True)
        )
        output_norm = self.output_norm_config.init(self.model_dim)
        return Decoder(
            self,
            embedding=embedding,
            global_rope=global_rope,
            local_rope=local_rope,
            layers=layers,
            output_norm=output_norm,
        )


class Decoder(LalamoModule[DecoderConfig]):
    embedding: EmbeddingBase
    global_rope: RoPE
    local_rope: RoPE | None
    layers: tuple[DecoderLayer, ...]
    output_norm: RMSNorm

    @property
    def activation_precision(self) -> DTypeLike:
        return self.embedding.activation_precision

    def __call__(
        self,
        token_ids: Int[Array, " suffix_tokens"],
        token_positions: Int[Array, " suffix_tokens"],
        kv_cache: KVCache | None = None,
        return_updated_kv_cache: bool = False,
        return_activation_trace: bool = False,
        length_without_padding: Int[Array, ""] | int | None = None,
    ) -> DecoderResult:
        maybe_kv_cache = kv_cache or ([None] * len(self.layers))
        inner_features = self.embedding.embed(token_ids)

        global_positional_embeddings = self.global_rope(token_positions)
        if self.local_rope is not None:
            local_positional_embeddings = self.local_rope(token_positions)
        else:
            local_positional_embeddings = global_positional_embeddings

        updated_kv_cache_layers = []
        layer_results = []
        for layer, kv_cache_slice in zip(self.layers, maybe_kv_cache, strict=True):
            if layer.attention_type == AttentionType.SLIDING_WINDOW:
                positional_embeddings_to_use = local_positional_embeddings
            else:
                positional_embeddings_to_use = global_positional_embeddings

            layer_result = layer(
                inner_features,
                positional_embeddings_to_use,
                kv_cache=kv_cache_slice,
                return_updated_kv_cache=return_updated_kv_cache,
                return_activation_trace=return_activation_trace,
                length_without_padding=length_without_padding,
            )
            inner_features = layer_result.outputs
            layer_results.append(layer_result)
            updated_kv_cache_layers.append(layer_result.updated_kv_cache)

        normalized_outputs = vmap(self.output_norm, in_axes=0)(inner_features)
        logits = vmap(self.embedding.readout, in_axes=0)(normalized_outputs)

        if return_activation_trace:
            activation_trace = DecoderActivationTrace(
                token_ids=token_ids,
                token_positions=token_positions,
                kv_cache=kv_cache,
                global_positional_embeddings=global_positional_embeddings,
                local_positional_embeddings=local_positional_embeddings,
                layer_results=tuple(layer_results),
                output_norm=normalized_outputs,
            )
        else:
            activation_trace = None

        if return_updated_kv_cache:
            updated_kv_cache = KVCache(updated_kv_cache_layers)
        else:
            updated_kv_cache = None

        return DecoderResult(
            logits=logits,
            updated_kv_cache=updated_kv_cache,
            activation_trace=activation_trace,
        )

    def init_static_kv_cache(self, capacity: int) -> KVCache:
        return KVCache(layer.init_static_kv_cache(capacity) for layer in self.layers)

    def export_weights(self, weight_layout: WeightLayout = WeightLayout.AUTO) -> ParameterDict:
        result = ParameterDict(
            embedding=self.embedding.export_weights(weight_layout),
            global_rope=self.global_rope.export_weights(weight_layout),
            layers=[layer.export_weights(weight_layout) for layer in self.layers],
            output_norm=self.output_norm.export_weights(weight_layout),
        )
        if self.local_rope:
            result["local_rope"] = self.local_rope.export_weights(weight_layout)
        return result



from abc import abstractmethod
from dataclasses import dataclass

import jax
import jax.numpy as jnp
from einops import rearrange
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterDict
from lalamo.quantization import QuantizationMode, dynamically_quantize_activations, quantize_weights

from .common import LalamoModule, WeightLayout, register_config_union
from .utils import apply_soft_capping

__all__ = [
    "EmbeddingBase",
    "EmbeddingConfig",
    "QuantizedTiedEmbedding",
    "QuantizedTiedEmbeddingConfig",
    "TiedEmbedding",
    "TiedEmbeddingConfig",
    "UntiedEmbedding",
    "UntiedEmbeddingConfig",
]


@dataclass(frozen=True)
class EmbeddingConfigBase:
    input_scale: float | None
    logits_soft_cap: float | None

    @abstractmethod
    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "EmbeddingBase": ...


class EmbeddingBase[ConfigT: EmbeddingConfigBase](LalamoModule[ConfigT]):
    @abstractmethod
    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]: ...

    @abstractmethod
    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]: ...

    @property
    @abstractmethod
    def vocab_size(self) -> int: ...

    @property
    @abstractmethod
    def model_dim(self) -> int: ...

    @classmethod
    def _default_weight_layout(cls) -> WeightLayout:
        return WeightLayout.INPUT_OUTPUT

    def embed(self, x: Int[Array, " tokens"]) -> Float[Array, "tokens channels"]:
        result = self._prepare_input_weights()[x]
        if self.config.input_scale is not None:
            result = result * jnp.array(self.config.input_scale, dtype=result.dtype)
        return result

    def readout(self, x: Float[Array, " channels"]) -> Float[Array, " vocabulary"]:
        logits = self._prepare_output_weights() @ x
        if self.config.logits_soft_cap is not None:
            logits = apply_soft_capping(logits, self.config.logits_soft_cap)
        return logits


@dataclass(frozen=True)
class TiedEmbeddingConfig(EmbeddingConfigBase):
    precision: DTypeLike

    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "TiedEmbedding":
        weights = jax.random.normal(key, (vocab_size, model_dim), dtype=self.precision)
        return TiedEmbedding(config=self, weights=weights)


class TiedEmbedding(EmbeddingBase[TiedEmbeddingConfig]):
    weights: Float[Array, "vocabulary channels"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    def __post_init__(self) -> None:
        if self.config.precision != self.weights.dtype:
            raise ValueError(
                f"Embedding dtype {self.weights.dtype} does not match the specified precision {self.config.precision}",
            )

    @property
    def model_dim(self) -> int:
        _, model_dim = self.weights.shape
        return model_dim

    @property
    def vocab_size(self) -> int:
        vocab_size, _ = self.weights.shape
        return vocab_size

    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]:
        return self.weights

    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]:
        return self.weights

    def export_weights(self, weight_layout: WeightLayout = WeightLayout.AUTO) -> ParameterDict:  # noqa: ARG002
        return ParameterDict(weights=self.weights)


@dataclass(frozen=True)
class UntiedEmbeddingConfig(EmbeddingConfigBase):
    precision: DTypeLike

    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "UntiedEmbedding":
        input_key, output_key = jax.random.split(key)
        input_weights = jax.random.normal(input_key, (vocab_size, model_dim), dtype=self.precision)
        output_weights = jax.random.normal(output_key, (vocab_size, model_dim), dtype=self.precision)
        return UntiedEmbedding(
            config=self,
            input_weights=input_weights,
            output_weights=output_weights,
        )


class UntiedEmbedding(EmbeddingBase[UntiedEmbeddingConfig]):
    input_weights: Float[Array, "vocabulary channels"]
    output_weights: Float[Array, "vocabulary channels"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def model_dim(self) -> int:
        _, model_dim = self.input_weights.shape
        return model_dim

    @property
    def vocab_size(self) -> int:
        vocab_size, _ = self.input_weights.shape
        return vocab_size

    def __post_init__(self) -> None:
        if self.config.precision != self.input_weights.dtype:
            raise ValueError(
                f"Embedding dtype {self.input_weights.dtype} does not match",
                f" the specified precision {self.config.precision}",
            )
        if self.config.precision != self.output_weights.dtype:
            raise ValueError(
                f"Embedding dtype {self.output_weights.dtype} does not match"
                f" the specified precision {self.config.precision}",
            )
        input_vocab_size, input_model_dim = self.input_weights.shape
        output_vocab_size, output_model_dim = self.output_weights.shape
        if input_vocab_size != output_vocab_size:
            raise ValueError(
                f"Input vocab size {input_vocab_size} does not match the output vocab size {output_vocab_size}",
            )
        if input_model_dim != output_model_dim:
            raise ValueError(
                f"Input model dim {input_model_dim} does not match the output model dim {output_model_dim}",
            )

    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]:
        return self.input_weights

    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]:
        return self.output_weights

    def export_weights(self, weight_layout: WeightLayout = WeightLayout.AUTO) -> ParameterDict:
        if weight_layout == WeightLayout.AUTO:
            weight_layout = self._default_weight_layout()

        match weight_layout:
            case WeightLayout.OUTPUT_INPUT:
                output_weights = self.output_weights
            case WeightLayout.INPUT_OUTPUT:
                output_weights = rearrange(self.output_weights, "token_ids channels -> channels token_ids")
            case _:
                raise ValueError(f"Unsupported weight layout: {weight_layout}")

        return ParameterDict(
            input_weights=self.input_weights,
            output_weights=output_weights,
        )


@dataclass(frozen=True)
class QuantizedTiedEmbeddingConfig(EmbeddingConfigBase):
    embedding_quantization_mode: QuantizationMode
    activation_quantization_mode: QuantizationMode | None
    activation_precision: DTypeLike

    def random_init(
        self,
        vocab_size: int,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "QuantizedTiedEmbedding":
        min_val, max_val = self.embedding_quantization_mode.range
        min_abs_val = min(abs(min_val), abs(max_val))
        scale = 1 / min_abs_val
        scales = scale * jnp.ones(vocab_size, dtype=self.activation_precision)
        weights = jax.random.normal(key, (vocab_size, model_dim), dtype=self.activation_precision)
        weights = quantize_weights(weights * min_abs_val, self.embedding_quantization_mode)
        return QuantizedTiedEmbedding(config=self, weights=weights, scales=scales)


class QuantizedTiedEmbedding(EmbeddingBase[QuantizedTiedEmbeddingConfig]):
    weights: Float[Array, "vocabulary channels"]
    scales: Float[Array, " vocabulary"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.activation_precision

    @property
    def model_dim(self) -> int:
        _, model_dim = self.weights.shape
        return model_dim

    @property
    def vocab_size(self) -> int:
        vocab_size, _ = self.weights.shape
        return vocab_size

    def __post_init__(self) -> None:
        if self.weights.dtype != self.config.activation_precision:
            raise ValueError(
                f"Embedding dtype ({self.scales.dtype}) is not equal to specified activation precision"
                f" ({self.config.activation_precision})."
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        if self.scales.dtype != self.config.activation_precision:
            raise ValueError(
                f"Scales dtype {self.scales.dtype} does not match the specified activation precision"
                f" {self.config.activation_precision}"
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        weights_vocab_size, weights_model_dim = self.weights.shape
        (scales_vocab_size,) = self.scales.shape
        if weights_vocab_size != scales_vocab_size:
            raise ValueError(
                f"Embedding vocab size {weights_vocab_size} does not match"
                f" the scales dimension size {scales_vocab_size}",
            )

    @property
    def int_weights(self) -> Int[Array, "vocabulary channels"]:
        result = quantize_weights(self.weights, self.config.embedding_quantization_mode)
        return result.astype(self.config.embedding_quantization_mode.dtype)

    def _prepare_weights(self) -> Float[Array, "vocabulary channels"]:
        quantized_weights = quantize_weights(self.weights, self.config.embedding_quantization_mode)
        quantized_weights = quantized_weights * self.scales.reshape(-1, 1)
        return quantized_weights

    def _prepare_input_weights(self) -> Float[Array, "vocabulary channels"]:
        return self._prepare_weights()

    def _prepare_output_weights(self) -> Float[Array, "vocabulary channels"]:
        return self._prepare_weights()

    def readout(self, x: Float[Array, " channels"]) -> Float[Array, " vocabulary"]:
        if self.config.activation_quantization_mode is not None:
            x = dynamically_quantize_activations(x, self.config.activation_quantization_mode)
        return super().readout(x)

    def export_weights(self, weight_layout: WeightLayout = WeightLayout.AUTO) -> ParameterDict:  # noqa: ARG002
        return ParameterDict(
            weights=self.int_weights,
            scales=self.scales,
        )


EmbeddingConfig = TiedEmbeddingConfig | UntiedEmbeddingConfig | QuantizedTiedEmbeddingConfig


register_config_union(EmbeddingConfig)



import importlib.metadata
from dataclasses import dataclass
from pathlib import Path
from typing import NamedTuple

import huggingface_hub
import jax.numpy as jnp
from jaxtyping import DTypeLike

from lalamo.modules import Decoder, DecoderConfig
from lalamo.quantization import QuantizationMode

from .model_specs import REPO_TO_MODEL, ModelSpec, UseCase

__all__ = [
    "REPO_TO_MODEL",
    "ModelMetadata",
    "ModelSpec",
    "import_model",
]


LALAMO_VERSION = importlib.metadata.version("lalamo")


@dataclass(frozen=True)
class ModelMetadata:
    toolchain_version: str
    vendor: str
    family: str
    name: str
    size: str
    quantization: QuantizationMode | None
    repo: str
    use_cases: tuple[UseCase, ...]
    model_config: DecoderConfig
    tokenizer_file_names: tuple[str, ...]


def download_weights(model_spec: ModelSpec, output_dir: Path | str | None = None) -> list[Path]:
    result = [
        huggingface_hub.hf_hub_download(
            repo_id=model_spec.repo,
            local_dir=output_dir,
            filename=filename,
        )
        for filename in model_spec.weights_file_names
    ]
    return [Path(path) for path in result]


def download_config_file(model_spec: ModelSpec, output_dir: Path | str | None = None) -> Path:
    result = huggingface_hub.hf_hub_download(
        repo_id=model_spec.repo,
        local_dir=output_dir,
        filename=model_spec.config_file_name,
    )
    return Path(result)


def download_tokenizer_files(model_spec: ModelSpec, output_dir: Path | str | None = None) -> tuple[Path, ...]:
    result = [
        huggingface_hub.hf_hub_download(
            repo_id=tokenizer_file_spec.repo or model_spec.repo,
            local_dir=output_dir,
            filename=tokenizer_file_spec.filename,
        )
        for tokenizer_file_spec in model_spec.tokenizer_files
    ]
    return tuple(Path(path) for path in result)


class ImportResults(NamedTuple):
    model: Decoder
    metadata: ModelMetadata
    tokenizer_file_paths: tuple[Path, ...]


def import_model(
    model_spec: ModelSpec,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
) -> ImportResults:
    foreign_config_file = download_config_file(model_spec)
    foreign_config = model_spec.config_type.from_json(foreign_config_file)

    tokenizer_file_paths = download_tokenizer_files(model_spec)
    if precision is None:
        precision = foreign_config.default_precision

    weights_paths = download_weights(model_spec)
    weights_dict = {}
    for weights_path in weights_paths:
        weights_dict.update(model_spec.weights_type.load(weights_path, precision))

    model = foreign_config.load_model(context_length, precision, accumulation_precision, weights_dict)
    metadata = ModelMetadata(
        toolchain_version=LALAMO_VERSION,
        vendor=model_spec.vendor,
        family=model_spec.family,
        name=model_spec.name,
        size=model_spec.size,
        quantization=model_spec.quantization,
        repo=model_spec.repo,
        use_cases=model_spec.use_cases,
        model_config=model.config,
        tokenizer_file_names=tuple(p.name for p in tokenizer_file_paths),
    )
    return ImportResults(model, metadata, tokenizer_file_paths)



from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import jax.numpy as jnp
from jaxtyping import Array, DTypeLike
from safetensors.flax import load_file as load_safetensors

from lalamo.model_import.configs import ForeignConfig
from lalamo.quantization import QuantizationMode
from lalamo.utils import torch_to_jax

__all__ = [
    "HUGGINFACE_GENERATION_CONFIG_FILE",
    "HUGGINGFACE_TOKENIZER_FILES",
    "ModelSpec",
    "TokenizerFileSpec",
    "UseCase",
    "awq_model_spec",
    "build_quantized_models",
    "huggingface_weight_files",
]


def cast_if_float(array: Array, cast_to: DTypeLike) -> Array:
    if array.dtype in [jnp.float16, jnp.bfloat16, jnp.float32, jnp.float64]:
        return array.astype(cast_to)
    return array


class WeightsType(Enum):
    SAFETENSORS = "safetensors"
    TORCH = "torch"

    def load(self, filename: Path | str, float_dtype: DTypeLike) -> dict[str, jnp.ndarray]:
        if self == WeightsType.SAFETENSORS:
            return {k: cast_if_float(v, float_dtype) for k, v in load_safetensors(filename).items()}

        import torch

        torch_weights = torch.load(filename, map_location="cpu", weights_only=True)
        return {k: cast_if_float(torch_to_jax(v), float_dtype) for k, v in torch_weights.items()}


class UseCase(Enum):
    CODE = "code"


@dataclass(frozen=True)
class TokenizerFileSpec:
    repo: str | None
    filename: str


@dataclass(frozen=True)
class ModelSpec:
    vendor: str
    family: str
    name: str
    size: str
    quantization: QuantizationMode | None
    repo: str
    config_type: type[ForeignConfig]
    config_file_name: str
    weights_file_names: tuple[str, ...]
    weights_type: WeightsType
    tokenizer_files: tuple[TokenizerFileSpec, ...] = tuple()
    use_cases: tuple[UseCase, ...] = tuple()


def huggingface_weight_files(num_shards: int) -> tuple[str, ...]:
    if num_shards == 1:
        return ("model.safetensors",)
    return tuple(f"model-{i:05d}-of-{num_shards:05d}.safetensors" for i in range(1, num_shards + 1))


def awq_model_spec(
    model_spec: ModelSpec,
    repo: str,
    quantization: QuantizationMode = QuantizationMode.UINT4,
) -> ModelSpec:
    return ModelSpec(
        vendor=model_spec.vendor,
        family=model_spec.family,
        name=f"{model_spec.name}-AWQ",
        size=model_spec.size,
        quantization=quantization,
        repo=repo,
        config_type=model_spec.config_type,
        config_file_name=model_spec.config_file_name,
        weights_file_names=huggingface_weight_files(1),
        weights_type=model_spec.weights_type,
        tokenizer_files=model_spec.tokenizer_files,
        use_cases=model_spec.use_cases,
    )


def build_quantized_models(model_specs: list[ModelSpec]):
    quantization_compatible_repos: list[str] = [
        "Qwen/Qwen2.5-3B-Instruct",
        "Qwen/Qwen2.5-7B-Instruct",
        "Qwen/Qwen2.5-Coder-3B-Instruct",
        "Qwen/Qwen2.5-Coder-7B-Instruct",
        "deepseek-ai/DeepSeek-R1-Distill-Qwen-1.5B",
        "HuggingFaceTB/SmolLM2-1.7B-Instruct",
        "meta-llama/Llama-3.2-3B-Instruct",
    ]

    quantized_model_specs: list[ModelSpec] = []
    for model_spec in model_specs:
        if model_spec.repo not in quantization_compatible_repos:
            continue
        quantized_repo = "trymirai/{}-AWQ".format(model_spec.repo.split("/")[-1])
        quantized_model_spec = awq_model_spec(model_spec, quantized_repo)
        quantized_model_specs.append(quantized_model_spec)
    return quantized_model_specs


HUGGINGFACE_TOKENIZER_FILES = (
    TokenizerFileSpec(repo=None, filename="tokenizer.json"),
    TokenizerFileSpec(repo=None, filename="tokenizer_config.json"),
)

HUGGINFACE_GENERATION_CONFIG_FILE = TokenizerFileSpec(repo=None, filename="generation_config.json")



from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from types import UnionType

import equinox as eqx
from cattrs import Converter
from jax import numpy as jnp
from jaxtyping import DTypeLike

from lalamo.common import ParameterDict

__all__ = [
    "AttentionType",
    "DummyUnionMember",
    "LalamoModule",
    "config_converter",
    "register_config_union",
]


class WeightLayout(Enum):
    AUTO = "auto"
    INPUT_OUTPUT = "input_output"
    OUTPUT_INPUT = "output_input"

    def __str__(self) -> str:
        match self:
            case WeightLayout.AUTO:
                return "auto"
            case WeightLayout.INPUT_OUTPUT:
                return "(input, output)"
            case WeightLayout.OUTPUT_INPUT:
                return "(output, input)"


class AttentionType(Enum):
    GLOBAL = "global"
    SLIDING_WINDOW = "sliding_window"


class LalamoModule[ConfigT](eqx.Module):
    config: ConfigT = eqx.field(static=True)

    @property
    @abstractmethod
    def activation_precision(self) -> DTypeLike: ...

    @abstractmethod
    def export_weights(self, weight_layout: WeightLayout = WeightLayout.AUTO) -> ParameterDict: ...


def _dtype_to_str(dtype: DTypeLike) -> str:
    if dtype == jnp.bfloat16:
        return "bfloat16"
    try:
        return str(dtype.dtype)  # type: ignore
    except AttributeError:
        return str(dtype)


def _str_to_dtype(dtype_str: str) -> jnp.dtype:
    return {
        "int4": jnp.int4,
        "int8": jnp.int8,
        "int16": jnp.int16,
        "int32": jnp.int32,
        "int64": jnp.int64,
        "bfloat16": jnp.bfloat16,
        "float16": jnp.float16,
        "float32": jnp.float32,
        "float64": jnp.float64,
    }[dtype_str]


config_converter = Converter()


config_converter.register_unstructure_hook_func(
    lambda t: t in [jnp.dtype, DTypeLike],
    _dtype_to_str,
)

config_converter.register_structure_hook_func(
    lambda t: t in [jnp.dtype, DTypeLike],
    lambda s, _: _str_to_dtype(s),
)


def register_config_union(union_type: UnionType) -> None:
    union_members = union_type.__args__
    name_to_type = {m.__name__: m for m in union_members}

    def unstructure(obj: object) -> dict | None:
        if obj is None:
            return None
        return {
            "type": obj.__class__.__name__,
            **config_converter.unstructure(obj),
        }

    config_converter.register_unstructure_hook(
        union_type,
        unstructure,
    )

    config_converter.register_unstructure_hook(
        union_type | None,
        unstructure,
    )

    def structure[T](config: dict | None, _: type[T]) -> T | None:
        if config is None:
            return None
        new_config = dict(config)
        type_name = new_config.pop("type")
        target_type = name_to_type[type_name]
        return name_to_type[type_name](**config_converter.structure(new_config, target_type))

    config_converter.register_structure_hook(
        union_type,
        structure,
    )

    config_converter.register_structure_hook(
        union_type | None,
        structure,
    )


@dataclass
class DummyUnionMember:
    pass



import jax
from jaxtyping import Array, Float

__all__ = [
    "apply_soft_capping",
]


def apply_soft_capping(
    values: Float[Array, "*"],
    soft_cap: float,
) -> Float[Array, "dst_tokens src_tokens"]:
    return jax.nn.tanh(values / soft_cap) * soft_cap



from collections.abc import Iterable, Mapping

import jax.numpy as jnp
from jaxtyping import Array, DTypeLike

__all__ = [
    "DEFAULT_PRECISION",
    "ParameterDict",
    "ParameterPath",
]

DEFAULT_PRECISION: DTypeLike = jnp.bfloat16


type NestedParameters = Mapping[str, Array | NestedParameters] | Iterable[Array | NestedParameters]


class ParameterDict(dict[str, Array]):
    def __init__(self, **kwargs: Array | NestedParameters | Iterable[Array | NestedParameters]) -> None:
        super().__init__(self._flatten(kwargs))

    def __setitem__(
        self,
        key: str,
        value: Array | NestedParameters | Iterable[Array | NestedParameters],
    ) -> None:
        key = ParameterPath(key)

        if isinstance(value, Array):
            super().__setitem__(key, value)
            return

        for subkey, subvalue in self._flatten(value).items():
            super().__setitem__(key / subkey, subvalue)

    @classmethod
    def _flatten(cls, nested_parameters: NestedParameters) -> dict[str, Array]:
        result: dict[str, Array] = {}
        if not isinstance(nested_parameters, Mapping):
            nested_parameters = {str(i): value for i, value in enumerate(nested_parameters)}
        for key, value in nested_parameters.items():
            key_path = ParameterPath(key)
            if isinstance(value, Array):
                result[key_path] = value
            else:
                result.update({key_path / subkey: subvalue for subkey, subvalue in cls._flatten(value).items()})
        return result


class ParameterPath(str):
    __slots__ = ()

    @property
    def components(self) -> tuple[str, ...]:
        return tuple(self.split("."))

    def __truediv__(self, other: str | int) -> "ParameterPath":
        if not self:
            return ParameterPath(str(other))
        return ParameterPath(self + "." + str(other))



from dataclasses import dataclass
from enum import Enum

import jax
import jax.numpy as jnp
import pytest
import torch

from lalamo import REPO_TO_MODEL, import_model
from tests.common import checkify_forward
from tests.huggingface_tracer import load_hf_tracer


class DType(Enum):
    FLOAT16 = "float16"
    BFLOAT16 = "bfloat16"
    FLOAT32 = "float32"

    @property
    def torch_dtype(self) -> torch.dtype:
        return getattr(torch, self.value)

    @property
    def jax_dtype(self) -> jnp.dtype:
        return jnp.dtype(self.value)


@dataclass
class Spec:
    model_repo: str
    dtype: DType
    requires_gpu: bool = False


MODEL_LIST = [
    Spec("Qwen/Qwen2.5-0.5B-Instruct", DType.FLOAT32),
    Spec("google/gemma-3-1b-it", DType.FLOAT32),
    Spec("deepseek-ai/DeepSeek-R1-Distill-Qwen-1.5B", DType.FLOAT32),
    Spec("meta-llama/Llama-3.2-1B-Instruct", DType.FLOAT32),
    Spec("PleIAs/Pleias-RAG-1B", DType.FLOAT32),
    Spec("Qwen/Qwen3-0.6B", DType.FLOAT32),
    Spec("Qwen/Qwen3-4B-AWQ", DType.FLOAT16, requires_gpu=True),
]


NUM_TOKENS = 512
TOKEN_STRIDE = 64


@pytest.fixture
def configure_precision_for_tests() -> None:
    jax.config.update("jax_default_matmul_precision", "highest")
    torch.backends.cudnn.allow_tf32 = False
    torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction = False
    torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction = False


@pytest.mark.parametrize("test_spec", MODEL_LIST)
def test_hf_model(test_spec: Spec, configure_precision_for_tests: None) -> None:  # noqa: ARG001
    if test_spec.requires_gpu and not torch.cuda.is_available():
        pytest.skip("GPU is required for this test")

    llm_model, *_ = import_model(
        REPO_TO_MODEL[test_spec.model_repo],
        context_length=NUM_TOKENS * TOKEN_STRIDE,
        precision=test_spec.dtype.jax_dtype,
    )
    hf_tracer = load_hf_tracer(test_spec.model_repo, torch_dtype=test_spec.dtype.torch_dtype)

    token_ids = jnp.arange(0, NUM_TOKENS, dtype=jnp.int32)
    token_positions = jnp.arange(0, NUM_TOKENS * TOKEN_STRIDE, TOKEN_STRIDE, dtype=jnp.int32)

    err, llm_result = checkify_forward(llm_model)(
        token_ids=token_ids,
        token_positions=token_positions,
        return_updated_kv_cache=True,
        return_activation_trace=True,
    )
    err.throw()

    hf_tracer.match_activations(llm_result)



from dataclasses import dataclass
from typing import NamedTuple

import equinox as eqx
import jax
from einops import einsum, rearrange, repeat
from jax import numpy as jnp
from jax import vmap
from jaxtyping import Array, Bool, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterDict
from lalamo.modules.normalization import RMSNorm, RMSNormConfig

from .common import AttentionType, LalamoModule, WeightLayout
from .kv_cache import DynamicKVCacheLayer, KVCacheLayer, StaticKVCacheLayer
from .linear import LinearBase, LinearConfig
from .rope import PositionalEmbeddings
from .utils import apply_soft_capping

__all__ = [
    "Attention",
    "AttentionConfig",
]


def _repeat_kv(
    keys_or_values: Float[Array, "tokens groups channels"],
    group_size: int,
) -> Float[Array, "tokens groups*group_size channels"]:
    return repeat(
        keys_or_values,
        "tokens groups channels -> tokens (groups group_size) channels",
        group_size=group_size,
    )


def _soft_capped_attention_kernel(
    queries: Float[Array, "dst_tokens heads head_channels"],
    keys: Float[Array, "src_tokens groups head_channels"],
    values: Float[Array, "src_tokens groups head_channels"],
    mask: Bool[Array, "dst_tokens src_tokens"] | None,
    scale: float | None,
    logit_soft_cap: float,
) -> Float[Array, "dst_tokens heads head_channels"]:
    dst_length, num_heads, head_dim = queries.shape
    src_length, num_groups, _ = keys.shape
    if scale is None:
        scale = head_dim**-0.5
    group_size = num_heads // num_groups
    keys = _repeat_kv(keys, group_size)
    values = _repeat_kv(values, group_size)
    queries_head_first = rearrange(queries, "dst_tokens heads channels -> heads dst_tokens channels")
    keys_head_first = rearrange(keys, "src_tokens heads channels -> heads src_tokens channels")
    attention_logits = einsum(
        queries_head_first,
        keys_head_first,
        "heads dst_tokens channels, heads src_tokens channels -> heads dst_tokens src_tokens",
    )
    if mask is not None:
        attention_logits = jnp.where(mask, attention_logits, jnp.array(float("-inf"), dtype=attention_logits.dtype))

    attention_logits = attention_logits * scale
    attention_logits = apply_soft_capping(attention_logits, logit_soft_cap)
    attention_weights = jax.nn.softmax(attention_logits, axis=-1)
    return einsum(
        attention_weights,
        values,
        "heads dst_tokens src_tokens, src_tokens heads channels -> dst_tokens heads channels",
    )


class AttentionResult(NamedTuple):
    outputs: Float[Array, "suffix_tokens channels"]
    kv_cache: KVCacheLayer | None = None


@dataclass(frozen=True)
class AttentionConfig:
    qkv_projection_config: LinearConfig
    out_projection_config: LinearConfig

    query_norm_config: RMSNormConfig | None
    key_norm_config: RMSNormConfig | None

    logit_soft_cap: float | None
    has_qkv_biases: bool
    has_out_biases: bool

    def random_init(
        self,
        model_dim: int,
        num_heads: int,
        num_groups: int,
        head_dim: int,
        is_causal: bool,
        scale: float | None,
        sliding_window_size: int | None,
        *,
        key: PRNGKeyArray,
    ) -> "Attention":
        qkv_key, out_key = jax.random.split(key)
        qkv_projection = self.qkv_projection_config.random_init(
            input_dim=model_dim,
            output_dims=(
                num_heads * head_dim,
                num_groups * head_dim,
                num_groups * head_dim,
            ),
            has_biases=self.has_qkv_biases,
            key=qkv_key,
        )
        out_projection = self.out_projection_config.random_init(
            num_heads * head_dim,
            (model_dim,),
            has_biases=self.has_out_biases,
            key=out_key,
        )

        if self.query_norm_config is not None:
            query_norm = self.query_norm_config.init(
                channels=head_dim,
            )
        else:
            query_norm = None

        if self.key_norm_config is not None:
            key_norm = self.key_norm_config.init(
                channels=head_dim,
            )
        else:
            key_norm = None

        return Attention(
            self,
            qkv_projection=qkv_projection,
            out_projection=out_projection,
            query_norm=query_norm,
            key_norm=key_norm,
            num_heads=num_heads,
            num_groups=num_groups,
            head_dim=head_dim,
            is_causal=is_causal,
            scale=scale,
            sliding_window_size=sliding_window_size,
        )


class Attention(LalamoModule[AttentionConfig]):
    qkv_projection: LinearBase
    out_projection: LinearBase

    query_norm: RMSNorm | None
    key_norm: RMSNorm | None

    num_heads: int = eqx.field(static=True)
    num_groups: int = eqx.field(static=True)
    head_dim: int = eqx.field(static=True)

    is_causal: bool = eqx.field(static=True)

    scale: float | None = eqx.field(static=True)
    sliding_window_size: int | None = eqx.field(static=True)

    @property
    def activation_precision(self) -> DTypeLike:
        return self.qkv_projection.activation_precision

    @property
    def model_dim(self) -> int:
        return self.qkv_projection.input_dim

    @property
    def group_size(self) -> int:
        return self.num_heads // self.num_groups

    @property
    def use_sliding_window(self) -> bool:
        return self.sliding_window_size is not None

    @property
    def attention_type(self) -> AttentionType:
        return AttentionType.SLIDING_WINDOW if self.sliding_window_size is not None else AttentionType.GLOBAL

    def __post_init__(self) -> None:
        if self.qkv_projection.has_biases != self.config.has_qkv_biases:
            raise ValueError(
                f"QKV projection has_biases {self.qkv_projection.has_biases} does not match"
                f" the specified config has_qkv_biases {self.config.has_qkv_biases}",
            )
        if self.out_projection.has_biases != self.config.has_out_biases:
            raise ValueError(
                f"Output projection has_biases {self.out_projection.has_biases} does not match"
                f" the specified config has_out_biases {self.config.has_out_biases}",
            )
        if self.query_norm is not None and self.query_norm.input_dim != self.head_dim:
            raise ValueError(
                f"Query normalization input dimension must match head_dim ({self.head_dim}),"
                f" got {self.query_norm.input_dim}",
            )
        if self.key_norm is not None and self.key_norm.input_dim != self.head_dim:
            raise ValueError(
                f"Key normalization input dimension must match head_dim ({self.head_dim}),"
                f" got {self.key_norm.input_dim}",
            )
        if self.num_heads % self.num_groups != 0:
            raise ValueError(
                "Number of heads must be divisible by the number of groups,"
                f" got {self.num_heads} heads and {self.num_groups} groups",
            )
        if self.out_projection.input_dim != self.num_heads * self.head_dim:
            raise ValueError(
                f"Output projection input dimension must be num_heads * head_dim"
                f" ({self.num_heads} * {self.head_dim} = {self.num_heads * self.head_dim}),"
                f" got {self.out_projection.input_dim}",
            )
        q_output_dim, k_output_dim, v_output_dim = self.qkv_projection.output_dims
        if q_output_dim != self.num_heads * self.head_dim:
            raise ValueError(
                f"Query projection output dimension must be num_heads * head_dim"
                f" ({self.num_heads} * {self.head_dim} = {self.num_heads * self.head_dim}),"
                f" got {q_output_dim}",
            )
        if k_output_dim != self.num_groups * self.head_dim:
            raise ValueError(
                f"Key projection output dimension must be num_groups * head_dim"
                f" ({self.num_groups} * {self.head_dim} = {self.num_groups * self.head_dim}),"
                f" got {k_output_dim}",
            )
        if v_output_dim != self.num_groups * self.head_dim:
            raise ValueError(
                f"Value projection output dimension must be num_groups * head_dim"
                f" ({self.num_groups} * {self.head_dim} = {self.num_groups * self.head_dim}),"
                f" got {v_output_dim}",
            )

    def __call__(
        self,
        inputs: Float[Array, "suffix_tokens channels"],
        positional_embeddings: PositionalEmbeddings,
        kv_cache: KVCacheLayer | None = None,
        return_updated_kv_cache: bool = False,
        length_without_padding: Int[Array, ""] | int | None = None,
    ) -> AttentionResult:
        queries, keys, values = vmap(self.qkv_projection, in_axes=0)(inputs)
        queries = rearrange(
            queries,
            "tokens (heads head_channels) -> tokens heads head_channels",
            heads=self.num_heads,
            head_channels=self.head_dim,
        )
        keys = rearrange(
            keys,
            "tokens (groups head_channels) -> tokens groups head_channels",
            groups=self.num_groups,
            head_channels=self.head_dim,
        )
        values = rearrange(
            values,
            "tokens (groups head_channels) -> tokens groups head_channels",
            groups=self.num_groups,
            head_channels=self.head_dim,
        )

        if self.query_norm is not None:
            queries = vmap(vmap(self.query_norm))(queries)
        if self.key_norm is not None:
            keys = vmap(vmap(self.key_norm))(keys)

        apply_positional_embeddings = vmap(positional_embeddings.apply, in_axes=1, out_axes=1)
        queries = apply_positional_embeddings(queries)
        keys = apply_positional_embeddings(keys)

        if kv_cache is None:
            updated_kv_cache = DynamicKVCacheLayer.init(keys, values, length=length_without_padding)
        else:
            updated_kv_cache = kv_cache.extend(keys, values, added_length=length_without_padding)

        num_suffix_tokens, _, _ = queries.shape
        mask = updated_kv_cache.attention_mask(num_suffix_tokens, self.is_causal, self.sliding_window_size)

        if self.config.logit_soft_cap is not None:
            attention_output = _soft_capped_attention_kernel(
                queries,
                updated_kv_cache.keys,
                updated_kv_cache.values,
                mask=mask,
                scale=self.scale,
                logit_soft_cap=self.config.logit_soft_cap,
            )
        else:
            attention_output = jax.nn.dot_product_attention(
                queries,
                updated_kv_cache.keys,
                updated_kv_cache.values,
                mask=mask,
                scale=self.scale,
            )
        attention_output = rearrange(
            attention_output,
            "tokens heads head_channels -> tokens (heads head_channels)",
            heads=self.num_heads,
            head_channels=self.head_dim,
        )
        (result,) = vmap(self.out_projection, in_axes=0)(attention_output)

        if not return_updated_kv_cache:
            updated_kv_cache = None

        return AttentionResult(
            outputs=result,
            kv_cache=updated_kv_cache,
        )

    def init_static_kv_cache(self, capacity: int) -> StaticKVCacheLayer:
        return StaticKVCacheLayer.empty(capacity, self.num_groups, self.head_dim, self.activation_precision)

    def export_weights(self, weight_layout: WeightLayout = WeightLayout.AUTO) -> ParameterDict:
        result = ParameterDict(
            qkv_projection=self.qkv_projection.export_weights(weight_layout),
            out_projection=self.out_projection.export_weights(weight_layout),
        )
        if self.query_norm is not None:
            result["query_norm"] = self.query_norm.export_weights(weight_layout)
        if self.key_norm is not None:
            result["key_norm"] = self.key_norm.export_weights(weight_layout)
        return result



from abc import abstractmethod
from typing import Self

import equinox as eqx
import jax.numpy as jnp
from jax.lax import dynamic_update_slice_in_dim
from jax.tree_util import register_pytree_node_class
from jaxtyping import Array, Bool, DTypeLike, Float, Int

from lalamo.common import ParameterDict

__all__ = ["DynamicKVCacheLayer", "KVCache", "KVCacheLayer", "StaticKVCacheLayer"]


class KVCacheLayer(eqx.Module):
    keys: Float[Array, "tokens groups head_channels"]
    values: Float[Array, "tokens groups head_channels"]

    def __post_init__(self) -> None:
        if self.keys.ndim != 3:
            raise ValueError(
                f"Key and value buffers must have 3 dimensions: capacity, groups, head_channels,"
                f" got shape {self.keys.shape}",
            )
        if self.keys.shape != self.values.shape:
            raise ValueError("Keys and values buffers must have the same shape")
        if self.keys.dtype != self.values.dtype:
            raise ValueError("Keys and values buffers must have the same dtype")

    @abstractmethod
    def attention_mask(
        self,
        suffix_length: int,
        is_causal: bool,
        sliding_window_size: int | None = None,
    ) -> Bool[Array, "suffix_tokens tokens"]: ...

    @abstractmethod
    def extend(
        self,
        added_keys: Float[Array, "new_tokens groups head_channels"],
        added_values: Float[Array, "new_tokens groups head_channels"],
        added_length: Int[Array, ""] | int | None = None,
    ) -> Self: ...

    def export(self) -> ParameterDict:
        return ParameterDict(
            keys=self.keys,
            values=self.values,
        )


@register_pytree_node_class
class KVCache(tuple[KVCacheLayer, ...]):
    __slots__ = ()

    def tree_flatten(self) -> tuple[tuple[KVCacheLayer, ...], None]:
        return (tuple(self), None)

    @classmethod
    def tree_unflatten(cls, aux_data: None, children: tuple[KVCacheLayer, ...]) -> Self:  # noqa: ARG003
        return cls(children)


class DynamicKVCacheLayer(KVCacheLayer):
    padding_mask: Bool[Array, " tokens"] | None = None

    @classmethod
    def init(
        cls,
        keys: Float[Array, "tokens groups head_channels"],
        values: Float[Array, "tokens groups head_channels"],
        length: Int[Array, ""] | int | None = None,
    ) -> "DynamicKVCacheLayer":
        num_tokens, _, _ = keys.shape
        if length is None:
            padding_mask = None
        else:
            padding_mask = jnp.arange(num_tokens, dtype=jnp.int32) < length
        return cls(keys, values, padding_mask)

    def attention_mask(
        self,
        suffix_length: int,
        is_causal: bool,
        sliding_window_size: int | None = None,
    ) -> Bool[Array, "suffix_tokens tokens"]:
        total_num_tokens, _, _ = self.keys.shape
        result = jnp.ones((suffix_length, total_num_tokens), dtype=jnp.bool)
        if is_causal:
            result = jnp.tril(result, k=total_num_tokens - suffix_length)
        if sliding_window_size is not None:
            result = jnp.triu(result, k=1 - sliding_window_size)
        if self.padding_mask is not None:
            result = result & self.padding_mask[None, :]
        return result

    def extend(
        self,
        added_keys: Float[Array, "new_tokens groups head_channels"],
        added_values: Float[Array, "new_tokens groups head_channels"],
        added_length: Int[Array, ""] | int | None = None,
    ) -> "DynamicKVCacheLayer":
        updated_keys = jnp.concatenate([self.keys, added_keys], axis=0)
        updated_values = jnp.concatenate([self.values, added_values], axis=0)

        added_padded_length, _, _ = added_keys.shape
        if self.padding_mask is None and added_length is None:
            return DynamicKVCacheLayer(updated_keys, updated_values)
        if added_length is None:
            added_length = added_padded_length

        if self.padding_mask is not None:
            old_padding_mask = self.padding_mask
        else:
            old_num_tokens, _, _ = self.keys.shape
            old_padding_mask = jnp.ones(old_num_tokens, dtype=jnp.bool)

        added_padding_mask = jnp.arange(added_padded_length, dtype=jnp.int32) < added_length
        updated_padding_mask = jnp.concatenate([old_padding_mask, added_padding_mask], axis=0)
        return DynamicKVCacheLayer(updated_keys, updated_values, updated_padding_mask)


class StaticKVCacheLayer(KVCacheLayer):
    current_length: Int[Array, ""]

    def attention_mask(
        self,
        suffix_length: int,
        is_causal: bool,
        sliding_window_size: int | None = None,
    ) -> Bool[Array, "suffix_tokens tokens"]:
        if is_causal:
            query_offsets = jnp.arange(-suffix_length, 0, dtype=jnp.int32)
        else:
            query_offsets = jnp.zeros(suffix_length, dtype=jnp.int32)

        query_indices = self.current_length + query_offsets
        key_indices = jnp.arange(self.capacity, dtype=jnp.int32)

        result = query_indices[:, None] >= key_indices[None, :]
        if sliding_window_size is not None:
            swa_mask = query_indices[:, None] < (key_indices[None, :] + sliding_window_size)
            result = result & swa_mask

        return result

    @property
    def padding_mask(self) -> Bool[Array, " tokens"] | None:
        return jnp.arange(self.capacity, dtype=jnp.int32) < self.current_length

    @property
    def capacity(self) -> int:
        result, _, _ = self.keys.shape
        return result

    def extend(
        self,
        added_keys: Float[Array, "tokens groups head_channels"],
        added_values: Float[Array, "tokens groups head_channels"],
        added_length: Int[Array, ""] | int | None = None,
    ) -> "StaticKVCacheLayer":
        if added_keys.shape != added_values.shape:
            raise ValueError("Keys and values must have the same shape")
        num_added_tokens, new_num_groups, new_head_dim = added_keys.shape
        _, old_num_groups, old_head_dim = self.keys.shape
        if new_num_groups != old_num_groups or new_head_dim != old_head_dim:
            raise ValueError("New keys and values must have the same number of groups and head dimensions")

        if added_length is None:
            added_length = num_added_tokens

        updated_keys = dynamic_update_slice_in_dim(
            self.keys,
            added_keys,
            self.current_length,
            0,
            allow_negative_indices=False,
        )
        updated_values = dynamic_update_slice_in_dim(
            self.values,
            added_values,
            self.current_length,
            0,
            allow_negative_indices=False,
        )
        updated_sequence_length = self.current_length + added_length
        return StaticKVCacheLayer(keys=updated_keys, values=updated_values, current_length=updated_sequence_length)

    @classmethod
    def empty(cls, capacity: int, num_groups: int, head_dim: int, dtype: DTypeLike) -> Self:
        return cls(
            keys=jnp.empty((capacity, num_groups, head_dim), dtype=dtype),
            values=jnp.empty((capacity, num_groups, head_dim), dtype=dtype),
            current_length=jnp.array(0, dtype=jnp.int32),
        )



import json
import re
import shutil
import sys
from enum import Enum
from pathlib import Path
from typing import Annotated

import jax.numpy as jnp
import thefuzz.process
from click import Context as ClickContext
from click import Parameter as ClickParameter
from click import ParamType
from jaxtyping import DTypeLike
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table
from safetensors.flax import save_file
from typer import Argument, Exit, Option, Typer

from lalamo.model_import import REPO_TO_MODEL, ModelMetadata, ModelSpec, import_model
from lalamo.modules import WeightLayout, config_converter
from lalamo.utils import jax_uint4_to_packed_uint8

SCRIPT_NAME = Path(sys.argv[0]).name

DEFAULT_OUTPUT_DIR = Path("models")


class Precision(Enum):
    FLOAT32 = "float32"
    FLOAT16 = "float16"
    BFLOAT16 = "bfloat16"


console = Console()
err_console = Console(stderr=True)
app = Typer(
    rich_markup_mode="rich",
    add_completion=False,
    pretty_exceptions_show_locals=False,
)


class ModelParser(ParamType):
    name: str = "Huggingface Model Repo"

    def convert(self, value: str, param: ClickParameter | None, ctx: ClickContext | None) -> ModelSpec:
        result = REPO_TO_MODEL.get(value)
        if result is None:
            closest_repo = _closest_repo(value)
            error_message_parts = [
                f'"{value}".',
            ]
            if closest_repo:
                error_message_parts.append(
                    f' Perhaps you meant "{closest_repo}"?',
                )
            error_message_parts.append(
                f"\n\nUse the `{SCRIPT_NAME} list-models` command to see the list of currently supported models.",
            )
            error_message = "".join(error_message_parts)
            self.fail(error_message, param, ctx)
        return result


def _closest_repo(query: str, min_score: float = 80) -> str | None:
    if not REPO_TO_MODEL:
        return None
    (closest_match, score), *_ = thefuzz.process.extract(query, list(REPO_TO_MODEL))
    if closest_match and score >= min_score:
        return closest_match
    return None


def _error(message: str) -> None:
    panel = Panel(message, box=box.ROUNDED, title="Error", title_align="left", border_style="red")
    err_console.print(panel)
    raise Exit(1)


def _pack_uint4_weights(weights: dict[str, jnp.ndarray]) -> dict[str, jnp.ndarray]:
    packed_weights = {}
    for key, value in weights.items():
        if value.dtype == jnp.uint4:
            packed_weights[key] = jax_uint4_to_packed_uint8(value)
        else:
            packed_weights[key] = value
    return packed_weights


@app.command(help="Convert the model for use with the Uzu inference engine.")
def convert(
    model_repo: Annotated[
        ModelSpec,
        Argument(
            help=(
                "Huggingface model repo. Example: [cyan]'meta-llama/Llama-3.2-1B-Instruct'[/cyan]."
                "\n\n\n\n"
                f"You can use the [cyan]`{SCRIPT_NAME} list-models`[/cyan] command to get a list of supported models."
            ),
            click_type=ModelParser(),
            show_default=False,
            metavar="MODEL_REPO",
            autocompletion=lambda: list(REPO_TO_MODEL),
        ),
    ],
    precision: Annotated[
        Precision | None,
        Option(
            help="Precision to use for activations and non-quantized weights.",
            show_default="Native precision of the model",
        ),
    ] = None,
    weight_layout: Annotated[
        WeightLayout | None,
        Option(
            help=(
                "Order of dimensions in the weights of linear layers."
                "\n\n\n\n"
                "If set to AUTO, the layout will depend on the model."
            ),
            show_default="auto",
        ),
    ] = None,
    output_dir: Annotated[
        Path | None,
        Option(
            help="Directory to save the converted model to.",
            show_default="Saves the converted model in the `models/<model_name>` directory",
        ),
    ] = None,
    context_length: Annotated[
        int | None,
        Option(
            help="Maximum supported context length. Used to precompute positional embeddings.",
            show_default="Model's native maximum context length.",
        ),
    ] = None,
    include_traces: Annotated[
        bool,
        Option(
            help="Export activation traces for debugging purposes.",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Option(
            help="Overwrite existing model files.",
        ),
    ] = False,
) -> None:
    if precision is not None:
        precision_dtype = config_converter.structure(precision.value, DTypeLike)  # type: ignore
    else:
        precision_dtype = None

    if weight_layout is not None:
        weight_layout = WeightLayout(weight_layout)
    else:
        weight_layout = WeightLayout.AUTO

    if output_dir is None:
        output_dir = DEFAULT_OUTPUT_DIR / model_repo.name

    console.print(f"🚀 Converting [cyan]{model_repo.name}[/cyan] by [cyan]{model_repo.vendor}[/cyan].")
    conversion_strs = [
        f"⚙️ Using weight layout [cyan]{weight_layout}[/cyan]",
    ]
    if precision is not None:
        conversion_strs.append(
            f" and converting floating-point weights into [cyan]{precision.name.lower()}[/cyan] precision",
        )
    conversion_strs.append(".")
    console.print("".join(conversion_strs))

    if output_dir.exists() and not overwrite:
        answer = console.input(
            rf"⚠️ Output directory [cyan]{output_dir}[/cyan] already exists."
            r" Do you want to overwrite it? [cyan]\[y/n][/cyan]: ",
        )
        while answer.lower() not in ["y", "n", "yes", "no"]:
            answer = console.input("Please enter 'y' or 'n': ")
        if answer.lower() in ["y", "yes"]:
            shutil.rmtree(output_dir)
        else:
            console.print("Exiting...")
            raise Exit

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task("👨‍🍳 Cooking...")
        model, metadata, tokenizer_file_paths = import_model(
            model_repo,
            precision=precision_dtype,
            context_length=context_length,
        )
        progress.add_task(f"💾 Saving the model to {output_dir}")
        output_dir.mkdir(parents=True, exist_ok=True)

        weights = dict(model.export_weights(weight_layout))
        packed_weights = _pack_uint4_weights(weights)
        save_file(packed_weights, output_dir / "model.safetensors")

        config_json = config_converter.unstructure(metadata, ModelMetadata)
        with open(output_dir / "config.json", "w") as file:
            json.dump(config_json, file, indent=4)

        for path in tokenizer_file_paths:
            shutil.copy(path, output_dir / path.name)

        if include_traces:
            progress.add_task("🚁 Generating traces...")

            num_tokens = 512
            token_stride = 8
            token_ids = jnp.arange(0, num_tokens, dtype=jnp.int32)
            token_positions = jnp.arange(0, num_tokens * token_stride, token_stride, dtype=jnp.int32)
            result = model(
                token_ids,
                token_positions,
                return_updated_kv_cache=True,
                return_activation_trace=True,
            )
            traces = dict(result.export())
            save_file(traces, output_dir / "traces.safetensors")

    console.print(f"🧑‍🍳 Model successfully cooked and saved to [cyan]`{output_dir}`[/cyan]!")


def _model_size_string_to_int(
    size_str: str,
    _regex: re.Pattern = re.compile(r"(?P<number>(\d+)(\.\d*)?)(?P<suffix>[KMBT])"),
) -> float:
    match = _regex.match(size_str)
    factors = {
        "K": 1024**1,
        "M": 1024**2,
        "B": 1024**3,
        "T": 1024**4,
    }
    if match:
        return float(match.group("number")) * factors[match.group("suffix")]
    raise ValueError(f"Invalid size string: {size_str}")


@app.command(help="List the supported models.")
def list_models(
    plain: Annotated[
        bool,
        Option(
            help="Only list repo names without fancy formatting.",
        ),
    ] = False,
) -> None:
    sorted_specs = sorted(
        REPO_TO_MODEL.values(),
        key=lambda spec: (
            spec.vendor.lower(),
            spec.family.lower(),
            _model_size_string_to_int(spec.size),
            spec.name.lower(),
        ),
    )

    if plain:
        for spec in sorted_specs:
            console.print(spec.repo)
        return

    table = Table(
        show_header=True,
        header_style="bold",
        show_lines=True,
        box=box.ROUNDED,
    )
    table.add_column("Vendor", justify="left", style="magenta")
    table.add_column("Family", justify="left", style="magenta", no_wrap=True)
    table.add_column("Size", justify="right", style="magenta")
    table.add_column("Quant", justify="left", style="magenta")
    table.add_column("Repo", justify="left", style="cyan", no_wrap=True)
    for spec in sorted_specs:
        table.add_row(
            spec.vendor,
            spec.family,
            spec.size,
            str(spec.quantization),
            spec.repo,
        )
    console.print(table)


if __name__ == "__main__":
    app()
